MODULE PZMatchImage EXPORTS Main;

(* Computes a grey scale image showing matches between two numeric chains *)
(* Last edited on 1999-11-08 05:54:15 by hcgl *)
 
IMPORT PZLRChain, PZImage;
IMPORT ParseParams, Process, FileRd, Wr, OSError, Thread, Math;

FROM Stdio IMPORT stderr;
FROM PZTypes IMPORT LONG, LONGS;
 
<* FATAL Wr.Failure, Thread.Alerted, OSError.E *>

TYPE
  Options = RECORD
      xChain: TEXT;  (* Input invariant chain file (without ".fcv") *)
      yChain: TEXT;  (* Input invariant chain file (without ".fcv") *)
      outFile: TEXT; (* Output ".pgm" file *)
  END;

 
PROCEDURE Main() =
  VAR d: LONG;
  BEGIN
    WITH
      o = GetOptions(),
      exr = ReadCurvatureChain(o.xChain),
      ex = exr.c^,
      eyr = ReadCurvatureChain(o.yChain),
      ey = eyr.c^,
      NEX = NUMBER(ex),
      NEY = NUMBER(ey),
      mImage = NEW(REF PZImage.T, NEY, NEX)^
    DO
      FOR j := 0 TO NEX-1 DO
        FOR i := 0 TO NEY-1 DO
          (* d := ComputeSqueezeDistance(-ey[NEY-1-i], ex[j]); *)
          d := ComputeAlphaDistance(-ey[NEY-1-i], ex[j]);
          mImage[i,j] := 255 - ROUND(255.0d0 * (1.0d0 - d)); 
        END;
      END;  
      PZImage.WritePGM(o.outFile, mImage);
    END (* with *)
  END Main;
  
PROCEDURE ReadCurvatureChain(name: TEXT): PZLRChain.ReadData  =
  BEGIN
    WITH fName = name & ".fcv" DO
      Wr.PutText(stderr, fName & "\n");
      RETURN PZLRChain.Read(FileRd.Open(fName), headerOnly := FALSE)
    END
  END ReadCurvatureChain;

<* UNUSED *> 
PROCEDURE NormalizedDistance(VAR d: LONGS; dMax : LONG;)=
  BEGIN
    IF dMax < 1.0d0   THEN dMax := 1.0d0  END;
    WITH N = NUMBER(d) DO 
      FOR i := 0 TO N-1 DO 
        d[i] := d[i] / dMax;
      END;  
    END
  END NormalizedDistance;

<*UNUSED*>
PROCEDURE ComputeSqueezeDistance(a, b: LONG): LONG =
  CONST Sigma = 2.0d0;
  BEGIN 
    WITH 
      d = (a - b)/Sigma
    DO 
      RETURN Math.exp(-d*d)
    END;
  END ComputeSqueezeDistance;


PROCEDURE ComputeAlphaDistance(a, b: LONG): LONG =
  CONST 
    Delta = 0.01d0; 
    
    Epsilon = 0.01d0;
  BEGIN 
    WITH 
      delta2 = Delta * Delta,
      epsilon2 = Epsilon * Epsilon,
      s = a + b, s2 = s*s,
      d = a - b, d2 = d*d
    DO 
      RETURN (delta2 * s2 + epsilon2) / (d2 + delta2 * s2 + epsilon2)  ;
    END;
  END ComputeAlphaDistance;

PROCEDURE GetOptions(): Options =
  VAR o: Options;
  BEGIN
    WITH
      pp = NEW(ParseParams.T).init(stderr)
    DO
      TRY
        pp.getKeyword("-xChain");
        o.xChain := pp.getNext();
        
        pp.getKeyword("-yChain");
        o.yChain := pp.getNext();
               
        pp.getKeyword("-outFile");
        o.outFile := pp.getNext();
        
       pp.finish();                                       
      EXCEPT                                                            
      | ParseParams.Error =>                                              
          Wr.PutText(stderr, "Usage: PZMatchImage \\\n");
          Wr.PutText(stderr, "  -xChain NAME \\\n");
          Wr.PutText(stderr, "  -yChain NAME \\\n");
          Wr.PutText(stderr, "  -outFile NAME \n");
        Process.Exit(1);
      END;
    END;
    RETURN o
    
  END GetOptions;

BEGIN
  Main()
END PZMatchImage.

(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
