MODULE PZUpdateSegs EXPORTS Main;
(* Last edited on 1999-09-03 07:01:04 by hcgl *)

(* Converts segment files from old to new format. *)

IMPORT ParseParams, Process, FileRd, FileWr;
IMPORT Wr, OSError, Thread, Stdio, Fmt, TextWr;
IMPORT PZSymbolChain, PZSegment;  

FROM Stdio IMPORT stderr;
FROM PZTypes IMPORT NAT, NATS, BOOLS;
 
<* FATAL Wr.Failure, Thread.Alerted, OSError.E *>

TYPE
  Options = RECORD
      input: TEXT;          (* Input candidate file name. *)
      band: NAT;            (* Input nominal filtering bandwidth. *)
      chainDir: TEXT;       (* Directory where to find chains. *)
      chainPrefix: TEXT;    (* Curve file name prefix. *)
      output: TEXT;         (* Output candidate file name. *)
      nCurves: NAT;         (* Total num of curves. *) 
  END;

PROCEDURE Main() =
  BEGIN
    WITH 
      o = GetOptions(),
      curves = GetAllCurves(o.nCurves)^,
      chData = PZSymbolChain.ReadAll(
        o.chainPrefix, o.band, ".cvc",
        sel := curves, headerOnly := TRUE,
        dir := o.chainDir
      ),
      m = GetSampleNumbers(chData.chain^)^,
      cData = PZSegment.ReadOld(FileRd.Open(o.input & ".seg"), m),
      seg = cData.s^,
      wr = FileWr.Open(o.output & ".seg")
    DO
      PZSegment.Write(wr, cData.cmt & UpdateComments(o), seg);
    END;
 END Main;          

      
PROCEDURE GetOptions(): Options =
  VAR o: Options;
  BEGIN
    WITH
      pp = NEW(ParseParams.T).init(stderr)
    DO
      TRY
        pp.getKeyword("-input");
        o.input := pp.getNext();
        
        IF pp.keywordPresent("-chainDir") THEN
          o.chainDir := pp.getNext()
        ELSE
          o.chainDir := "."
        END;
        
        pp.getKeyword("-chainPrefix");
        o.chainPrefix := pp.getNext();
        
        pp.getKeyword("-band");
        o.band := pp.getNextInt();
        
        pp.getKeyword("-nCurves");
        o.nCurves := pp.getNextInt();
        
        pp.getKeyword("-output");
        o.output := pp.getNext();
         
        pp.finish();                                       

      EXCEPT                                                            
      | ParseParams.Error =>                                              
          Wr.PutText(stderr, "Usage: PZUpdateSegs \\\n");
          Wr.PutText(stderr, "  -input NAME \\\n");
          Wr.PutText(stderr, "  [ -chainDir DIR ] -chainPrefix NAME \\\n");
          Wr.PutText(stderr, "  -band NUMBER \\\n");
          Wr.PutText(stderr, "  -nCurves NUMBER \\\n");
          Wr.PutText(stderr, "  -output NAME\n");
        Process.Exit(1);
      END;
    END;
    RETURN o
  END GetOptions;
 
PROCEDURE GetAllCurves(n: NAT): REF BOOLS =
  BEGIN
    WITH 
      rseen = NEW(REF BOOLS, n),
      seen = rseen^
    DO
      FOR k := 0 TO LAST(seen) DO seen[k] := TRUE  END;
      RETURN rseen
    END
  END GetAllCurves;
  
PROCEDURE GetSampleNumbers(READONLY ch: ARRAY OF PZSymbolChain.ReadData): 
  REF NATS  =
  BEGIN
    WITH 
      n = NUMBER(ch),
      rm = NEW(REF NATS, n), 
      m = rm^
    DO
      FOR k := 0 TO n-1 DO 
        m[k] := ch[k].samples
      END;
      RETURN rm
    END
  END GetSampleNumbers;


PROCEDURE UpdateComments(READONLY o: Options):TEXT=
  BEGIN 
    WITH 
      wr = NEW(TextWr.T).init()
    DO 
      Wr.PutText( wr, "  input =       " & o.input & "\n");
      Wr.PutText( wr, "  chainDir =    " & o.chainDir & "\n");
      Wr.PutText( wr, "  chainPrefix = " & o.chainPrefix & "\n");
      Wr.PutText( wr, "  band =        " & Fmt.Int(o.band) & "\n");
      Wr.PutText( wr, "  output =      " & o.output & "\n");
      Wr.PutText( wr, "PZUpdateSegs\n");
      RETURN(TextWr.ToText(wr))
    END (* DO *);
  END UpdateComments;
 
 
BEGIN
  Main()
END PZUpdateSegs.
(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
