MODULE PZCandidate;

(* Last edited on 1999-11-16 14:02:57 by hcgl *)

IMPORT Rd, Wr, Fmt, FGet, NGet, FPut, NPut, FileFmt, Thread;
IMPORT PZMatch, PZSegment;
FROM PZTypes IMPORT LONG, INT, NAT, NATS, BOOL, BOOLS;

(* FROM Stdio IMPORT stderr; *)
(* <* FATAL Wr.Failure, Thread.Alerted *> *)

PROCEDURE IsEmpty(READONLY c: T): BOOL =
  BEGIN
    RETURN c.seg[0].ns = 0 OR c.seg[1].ns = 0
  END IsEmpty;

PROCEDURE Expand(READONLY c: T; iniExtra, finExtra: ARRAY [0..1] OF INT): T =
  BEGIN
    WITH
      oldSteps = c.seg[0].ns + c.seg[1].ns - 2,
      newSeg = PZSegment.Pair{
        PZSegment.Expand(c.seg[0], iniExtra[0], finExtra[0]),
        PZSegment.Expand(c.seg[1], iniExtra[1], finExtra[1])
      },
      newSteps = newSeg[0].ns + newSeg[1].ns - 2
    DO
      RETURN T{
        seg := newSeg,
        mismatch := c.mismatch,
        length := c.length * FLOAT(newSteps, LONG)/FLOAT(oldSteps, LONG),
        matchedLength := c.matchedLength,
        pm := PZMatch.AdjustPacked(c.pm, c.seg, newSeg)
      }
    END
  END Expand;

PROCEDURE TrimAndPack(
    READONLY cand: T;
    READONLY mt: PZMatch.T;
    mismatch: LONG;
    length: LONG;
    matchedLength: LONG;
  ): T =
  VAR t: PZSegment.Pair;  (* Matched segments *)
  BEGIN
    WITH
      nm = NUMBER(mt),
      rpm = PZMatch.Pack(mt)
    DO
      FOR j := 0 TO 1 DO
        WITH
          sj = cand.seg[j],
          tj = t[j]
        DO
          tj := sj;
          tj.cvx := sj.cvx;
          tj.tot := sj.tot;
          IF sj.rev THEN
            tj.ini := (sj.ini + sj.ns - 1 - mt[nm-1][j]) MOD sj.tot
          ELSE
            tj.ini := (sj.ini + mt[0][j]) MOD sj.tot
          END;
          tj.ns := mt[nm-1][j] - mt[0][j] + 1;
          tj.rev := sj.rev
        END
      END;
      RETURN T{
          seg := t,
          mismatch := mismatch,
          length := length,
          matchedLength := matchedLength,
          pm := PZMatch.AdjustPacked(rpm, cand.seg, t)
        }
    END;
  END TrimAndPack;

CONST
  OldFileVersion = "97-02-03";
  NewFileVersion = "99-07-25";

PROCEDURE Write(
    wr :Wr.T;
    cmt: TEXT;
    READONLY c: List;
    lambda: LONG;
  ) =
  <* FATAL Thread.Alerted, Wr.Failure *>
  BEGIN
    WITH
      n = LAST(c)
    DO
      FileFmt.WriteHeader(wr, "PZCandidate.List", NewFileVersion);
      FileFmt.WriteComment(wr, cmt, '|');
      NPut.Int(wr, "candidates", NUMBER(c));        FPut.EOL(wr);
      NPut.LongReal(wr, "lambda", lambda);     FPut.EOL(wr);
      FOR i := 0 TO n  DO
        WITH ci = c[i] DO
          FOR j := 0 TO 1 DO
            WITH cij = ci.seg[j] DO
              FPut.Int(wr, cij.cvx); FPut.Space(wr);
              FPut.Int(wr, cij.tot); FPut.Space(wr);
              FPut.Int(wr, cij.ini); FPut.Space(wr);
              FPut.Int(wr, cij.ns); FPut.Space(wr);
              IF cij.rev THEN
                FPut.Char(wr, '-')
              ELSE
                FPut.Char(wr, '+')
              END;
              FPut.Space(wr);
            END
          END;
          FPut.LongReal(wr, ci.mismatch, prec := 4); FPut.Space(wr);
          FPut.LongReal(wr, ci.length, prec := 2); FPut.Space(wr);
          FPut.LongReal(wr, ci.matchedLength, prec := 2);
          IF ci.pm # NIL THEN
            FPut.Space(wr); FPut.Space(wr);
            PZMatch.WritePacked(wr, ci.pm^);
          END;
          FPut.EOL(wr);
        END;
      END;
      FileFmt.WriteFooter(wr, "PZCandidate.List");
      Wr.Flush(wr);
    END;
  END Write;

PROCEDURE Read(rd :Rd.T): ReadData  =
  VAR d: ReadData;
  <* FATAL Rd.EndOfFile, Rd.Failure, Thread.Alerted *>
  BEGIN
    FileFmt.ReadHeader(rd, "PZCandidate.List", NewFileVersion);
    d.cmt := FileFmt.ReadComment(rd, '|');
    WITH
      n = NGet.Int(rd, "candidates")
     DO
      FGet.EOL(rd);
      d.c := NEW(REF List, n);
      d.lambda := NGet.LongReal(rd, "lambda"); FGet.EOL(rd);
      FOR i := 0 TO n-1 DO
        FGet.Skip(rd, FGet.Spaces);
        WITH ci = d.c[i] DO
          FOR j := 0 TO 1 DO
            WITH cij = ci.seg[j] DO
              cij.cvx := FGet.Int(rd); FGet.Skip(rd, FGet.Spaces);
              cij.tot := FGet.Int(rd); FGet.Skip(rd, FGet.Spaces);
              cij.ini := FGet.Int(rd); FGet.Skip(rd, FGet.Spaces);
              cij.ns := FGet.Int(rd); FGet.Skip(rd, FGet.Spaces);
              WITH c = Rd.GetChar(rd) DO
                IF c = '-' THEN
                  cij.rev := TRUE
                ELSIF c = '+' THEN
                  cij.rev := FALSE
                ELSE
                  <* ASSERT FALSE *>
                END
              END;
              FGet.Skip(rd, FGet.Spaces);
            END
          END;
          ci.mismatch := FGet.LongReal(rd); FGet.Skip(rd, FGet.Spaces);
          ci.length := FGet.LongReal(rd); FGet.Skip(rd, FGet.Spaces);
          ci.matchedLength := FGet.LongReal(rd); FGet.Skip(rd, FGet.Spaces);
          WITH c = Rd.GetChar(rd) DO
            Rd.UnGetChar(rd);
            IF c = '\n' THEN
              ci.pm := NIL;
            ELSE
              ci.pm := PZMatch.ReadPacked(rd)
            END
          END;
          FGet.EOL(rd)
        END
      END;
      FileFmt.ReadFooter(rd, "PZCandidate.List");
      RETURN d;
    END;
  END Read;

PROCEDURE ReadOld(
    rd :Rd.T;
    READONLY m: NATS;
    rev: ARRAY [0..1] OF BOOL;
  ): ReadData  =
  VAR d: ReadData;
  <* FATAL Rd.EndOfFile, Rd.Failure, Thread.Alerted *>
  BEGIN
    FileFmt.ReadHeader(rd, "PZCandidate.List", OldFileVersion);
    d.cmt := FileFmt.ReadComment(rd, '|');
    WITH
      n = NGet.Int(rd, "candidates")
     DO
      FGet.EOL(rd);
      d.c := NEW(REF List, n);
      d.lambda := NGet.LongReal(rd, "lambda"); FGet.EOL(rd);
      FOR i := 0 TO n-1 DO
        FGet.Skip(rd, FGet.Spaces);
        WITH ci = d.c[i] DO
          FOR j := 0 TO 1 DO
            WITH cij = ci.seg[j] DO
              cij.cvx := FGet.Int(rd); FGet.Skip(rd, FGet.Spaces);
              cij.tot := m[cij.cvx];
              cij.ini := FGet.Int(rd); FGet.Skip(rd, FGet.Spaces);
              WITH fin = FGet.Int(rd) DO
                cij.ns := (fin - cij.ini + 1) MOD cij.tot
              END;
              FGet.Skip(rd, FGet.Spaces);
              cij.rev := rev[j]
            END
          END;
          ci.mismatch := FGet.LongReal(rd); FGet.Skip(rd, FGet.Spaces);
          ci.length := FGet.LongReal(rd); FGet.Skip(rd, FGet.Spaces);
          ci.matchedLength := FGet.LongReal(rd); FGet.Skip(rd, FGet.Spaces);
          WITH c = Rd.GetChar(rd) DO
            Rd.UnGetChar(rd);
            IF c = '\n' THEN
              ci.pm := NIL;
            ELSE
              ci.pm := PZMatch.ReadPacked(rd)
            END
          END;
          FGet.EOL(rd)
        END
      END;
      FileFmt.ReadFooter(rd, "PZCandidate.List");
      RETURN d;
    END;
  END ReadOld;

PROCEDURE GetCurves(READONLY cand: List): REF BOOLS =
  CONST MaxCurves = 10000;
  VAR mxc: NAT := 0; (* Max curve number in "cand" *)
  BEGIN
    WITH seen = NEW(REF BOOLS, MaxCurves)^ DO
      FOR k := 0 TO LAST(seen) DO seen[k] := FALSE END;
      FOR i := 0 TO LAST(cand) DO
        FOR j := 0 TO 1 DO
          WITH k = cand[i].seg[j].cvx DO
            IF NOT seen[k] THEN
              seen[k] := TRUE;  mxc := MAX(mxc, k)
            END;
          END
        END
      END;
      mxc := mxc + 1;
      WITH rb = NEW(REF BOOLS, mxc + 1) DO
        rb^ := SUBARRAY(seen, 0, mxc + 1);
        RETURN rb
      END
    END
  END GetCurves;

PROCEDURE Print (wr: Wr.T; READONLY cand: T; pairing: BOOL := TRUE) =
  <* FATAL Wr.Failure, Thread.Alerted *>

  PROCEDURE PrintHalf(j: [0..1]; seg: PZSegment.T) =
    BEGIN
      Wr.PutText(wr, "  ");
      Wr.PutText(wr, "seg[" & Fmt.Int(j) & "]: ");
      PZSegment.Print(wr, seg);
      Wr.PutText(wr, "\n");
    END PrintHalf;

  BEGIN
    PrintHalf(0, cand.seg[0]);
    PrintHalf(1, cand.seg[1]);
    IF pairing AND cand.pm # NIL THEN
      PZMatch.WritePacked(wr, cand.pm^);
      Wr.PutText(wr, "\n");
    END;
    Wr.PutText(wr,"mismatch = " & FLR(cand.mismatch, 6, 2));
    Wr.PutText(wr, " ");
    Wr.PutText(wr,"length = " & FLR(cand.length, 9, 4));
    Wr.PutText(wr, " ");
    Wr.PutText(wr,"matchedLength = " & FLR(cand.matchedLength, 9, 4));
    Wr.PutText(wr, "\n");
  END Print;

PROCEDURE Sort(VAR c: List; before: CompareProc) =
  VAR  q, r, mx: NAT;
       cc, smx: T;
  BEGIN
    WITH n = NUMBER(c) DO
      (* 1. Build heap with largest element at c[0] *)
      FOR  p := 0  TO n-1  DO
        cc := c[p];
        r := p;
        LOOP
          IF r = 0 THEN  EXIT  END;
          q := (r-1) DIV 2;
          IF before(c[q], cc)  THEN
            c[r] := c[q];
          ELSE
            EXIT
          END;
          r := q
        END;
        c[r] := cc;
      END;
      (* 2. Remove elements from heap and insert at end *)
      FOR  p := n-1  TO 1  BY -1  DO
        (* save c[p] *)
        cc := c[p];
        (* Move largest heap element to pos[p] *)
        c[p] := c[0];
        (* Insert cc in remaining heap, from root down  *)
        q := 0;
        LOOP
          c[q] := cc;
          r := 2*q+1;
          (* Find largest among cc, c[LEFT(q)], c[RIGHT(q)]  *)
          mx := q; smx := cc;
          IF r < p AND before(smx, c[r])  THEN
            mx := r; smx := c[r]
          END;
          INC(r);
          IF r < p AND before(smx, c[r])  THEN
            mx := r; smx := c[r]
          END;
          (*  See who won  *)
          IF mx = q THEN
            (*  Stop here  *)
            EXIT
          ELSE
            (*  Promote child and advance  *)
            c[q] := c[mx];
            q := mx
          END
        END
      END;
      (* Paranoid check: *)
      FOR i := 1 TO n-1 DO <* ASSERT NOT before(c[i],c[i-1]) *> END
    END
  END Sort;

PROCEDURE LexicallyBetter(READONLY a, b: T): BOOL =
  BEGIN
    IF a.seg[0].cvx # b.seg[0].cvx THEN RETURN a.seg[0].cvx < b.seg[0].cvx END;
    <* ASSERT a.seg[0].tot = b.seg[0].tot *>
    IF a.seg[1].cvx # b.seg[1].cvx THEN RETURN a.seg[1].cvx < b.seg[1].cvx END;
    <* ASSERT a.seg[1].tot = b.seg[1].tot *>
    IF a.seg[0].ini # b.seg[0].ini THEN RETURN a.seg[0].ini < b.seg[0].ini END;
    IF a.seg[1].ini # b.seg[1].ini THEN RETURN a.seg[1].ini < b.seg[1].ini END;
    IF a.seg[0].ns # b.seg[0].ns THEN RETURN a.seg[0].ns < b.seg[0].ns END;
    IF a.seg[1].ns # b.seg[1].ns THEN RETURN a.seg[1].ns < b.seg[1].ns END;
    IF a.mismatch # b.mismatch THEN RETURN a.mismatch < b.mismatch END;
    IF a.length # b.length THEN RETURN a.length > b.length END;
    IF a.matchedLength # b.matchedLength THEN RETURN a.matchedLength > b.matchedLength END;
    RETURN FALSE
  END LexicallyBetter;

PROCEDURE PairMismatchBetter(READONLY a, b: T): BOOL =
  BEGIN
    IF a.seg[0].cvx # b.seg[0].cvx THEN RETURN a.seg[0].cvx < b.seg[0].cvx END;
    <* ASSERT a.seg[0].tot = b.seg[0].tot *>
    IF a.seg[1].cvx # b.seg[1].cvx THEN RETURN a.seg[1].cvx < b.seg[1].cvx END;
    <* ASSERT a.seg[1].tot = b.seg[1].tot *>
    IF a.mismatch # b.mismatch THEN RETURN a.mismatch < b.mismatch END;
    IF a.length # b.length THEN RETURN a.length > b.length END;
    IF a.matchedLength # b.matchedLength THEN RETURN a.matchedLength > b.matchedLength END;
    IF a.seg[0].ns # b.seg[0].ns THEN RETURN a.seg[0].ns > b.seg[0].ns END;
    IF a.seg[1].ns # b.seg[1].ns THEN RETURN a.seg[1].ns > b.seg[1].ns END;
    IF a.seg[0].ini # b.seg[0].ini THEN RETURN a.seg[0].ini < b.seg[0].ini END;
    IF a.seg[1].ini # b.seg[1].ini THEN RETURN a.seg[1].ini < b.seg[1].ini END;
    RETURN FALSE
  END PairMismatchBetter;

PROCEDURE MismatchBetter(READONLY a, b: T): BOOL =
  (*
    Compares by "mismatch", breaking ties by other criteria
  *)
  BEGIN
    IF a.mismatch # b.mismatch THEN RETURN a.mismatch < b.mismatch END;
    IF a.length # b.length THEN RETURN a.length > b.length END;
    IF a.matchedLength # b.matchedLength THEN RETURN a.matchedLength > b.matchedLength END;
    IF a.seg[0].cvx # b.seg[0].cvx THEN RETURN a.seg[0].cvx < b.seg[0].cvx END;
    <* ASSERT a.seg[0].tot = b.seg[0].tot *>
    IF a.seg[1].cvx # b.seg[1].cvx THEN RETURN a.seg[1].cvx < b.seg[1].cvx END;
    <* ASSERT a.seg[1].tot = b.seg[1].tot *>
    IF a.seg[0].ini # b.seg[0].ini THEN RETURN a.seg[0].ini < b.seg[0].ini END;
    IF a.seg[1].ini # b.seg[1].ini THEN RETURN a.seg[1].ini < b.seg[1].ini END;
    IF a.seg[0].ns # b.seg[0].ns THEN RETURN a.seg[0].ns < b.seg[0].ns END;
    IF a.seg[1].ns # b.seg[1].ns THEN RETURN a.seg[1].ns < b.seg[1].ns END;
    RETURN FALSE
  END MismatchBetter;

PROCEDURE RemoveDuplicates(VAR cand: List; VAR nCands: NAT) =
  VAR nKept: NAT;
  BEGIN
    nKept := 1;
    FOR i := 1 TO nCands-1 DO
      IF (cand[i].seg[0] # cand[nKept-1].seg[0])
      OR (cand[i].seg[1] # cand[nKept-1].seg[1])
      THEN
        IF nKept # i THEN cand[nKept] := cand[i] END;
        INC(nKept)
      END
    END;
    nCands := nKept;
  END RemoveDuplicates;

PROCEDURE RemoveShort(VAR cand: List; VAR nCands: NAT; minSteps: NAT) =
  VAR nKept: NAT;
  BEGIN
    nKept := 1;
    FOR i := 1 TO nCands-1 DO
      IF (cand[i].seg[0].ns > minSteps)
      OR (cand[i].seg[1].ns > minSteps)
      THEN
        IF nKept # i THEN cand[nKept] := cand[i] END;
        INC(nKept)
      END
    END;
    nCands := nKept;
  END RemoveShort;

PROCEDURE PruneCandsPerPair(VAR cand: List; VAR nCands: NAT; maxCands: NAT) =
  (*
    Assumes candidates are sorted by "PairMismatchBetter", so that all
    candidates for the same curve pair are in adjacent positions,
    sorted by decreasing quality.
  *)
  VAR nKept: NAT := 0;  (* Number of candidates kept *)
      nPair: NAT;       (* Number of candidates of current curve pair *)
      k0Old, k1Old: NAT := LAST(NAT);
  BEGIN
    FOR i := 0 TO nCands-1 DO
      WITH
        k0 = cand[i].seg[0].cvx,
        k1 = cand[i].seg[1].cvx
      DO
        IF k0 # k0Old OR k1 # k1Old THEN
          nPair := 0; k0Old := k0; k1Old := k1
        END;
        INC(nPair);
        IF (nPair <= maxCands) THEN
          IF i > nKept THEN cand[nKept] := cand[i] END;
          INC(nKept)
        END
      END
    END;
    nCands := nKept;
  END PruneCandsPerPair;

PROCEDURE PruneCandsPerCurve(VAR cand: List; VAR nCands: NAT; maxCands: NAT) =
  VAR nKept: NAT := 0;  (* Number of candidates kept *)
  BEGIN
    WITH
      curves = GetCurves(SUBARRAY(cand, 0, nCands))^,
      nCurves = NUMBER(curves),
      ctCurve = NEW(REF NATS, nCurves)^
    DO
      FOR i := 0 TO nCurves-1 DO ctCurve[i] := 0 END;
      FOR i := 0 TO nCands-1 DO
        WITH
          k0 = cand[i].seg[0].cvx,
          ct0 = ctCurve[k0],
          k1 = cand[i].seg[1].cvx,
          ct1 = ctCurve[k1]
        DO
          INC(ct0); INC(ct1);
          IF (ct0 <= maxCands OR ct1 <= maxCands) THEN
            IF i > nKept THEN cand[nKept] := cand[i] END;
            INC(nKept)
          END
        END
      END;
      nCands := nKept;
    END;
  END PruneCandsPerCurve;

PROCEDURE MergeOverlaps(
    VAR cand: List; 
    VAR nCands: NAT;
    minSteps: NAT;
    maxAdjust: NAT;
  ) =
  VAR nKept: NAT;     (* Number of candidates kept. *)
      ini: NAT;       (* First kept candidate for this curve pair. *)
      ra, rb: REF PZMatch.T := NIL;
  BEGIN
    (* Copy the first candidate to the kept part of "cand": *)
    ini := 0;
    nKept := 0;
    (* Process the remaining candidates: *)
    FOR i := 0 TO nCands DO
      IF nKept > 0 THEN
        IF i = nCands
        OR cand[i].seg[0].cvx # cand[nKept-1].seg[0].cvx
        OR cand[i].seg[1].cvx # cand[nKept-1].seg[1].cvx THEN
          (* New pair; process all candidates of previous pair. *)
          (* Those candidates are "cand[ini..nKept-1]". *)
          MergePairOverlaps(cand,
            minSteps, maxAdjust,
            ini, (*IO*) nKept,
            (*WORK*) ra, rb
          );
          (* Start copying next pair: *)
          ini := nKept
        END
      END;
      (* Save candidate for later merging: *)
      IF i < nCands THEN
        IF i > nKept THEN cand[nKept] := cand[i] END;
        INC(nKept)
      END
    END;
    nCands := nKept;
  END MergeOverlaps;

PROCEDURE MergePairOverlaps(
    VAR cand: List;
    minSteps: NAT;
    maxAdjust: NAT;
    ini: NAT;       (* Actual index of first candidate in "cand" *)
    VAR lim: NAT;   (* Index of first non-candidate in "cand" *)
    (*WORK*) VAR ra, rb: REF PZMatch.T;
  ) =
  (*
    Given a bunch of candidates for the same curve pair, "cand[ini..lim-1]",
    removes all duplicates and joins all overlapping candidtes,
    decrementing "lim" accordingly. *)
  VAR r, s: NAT;
  BEGIN
    (* This code assumes that "dist(A, Join(B,C)) = MIN(dist(A,B), dist(A,C))". *)
    (* The assumption may not be precisely correct, but... *)
    r := ini;
    WHILE r < lim DO
      (* Invariant: "cand[0..r-1]" are well separated from all other cands. *)
      (* Try to condense "cand[r]" with the following ones: *)
      s := r + 1;
      WHILE s < lim DO
        IF Overlap(cand[r], cand[s], minSteps, maxAdjust, ra, rb) THEN
          cand[r] := Join(cand[r], cand[s]);
          (* Fill the hole left by "cand[s]": *)
          IF s < lim-1 THEN cand[s] := cand[lim-1] END;
          DEC(lim);
          (* Now start over again, since "cand[r]" changed: *)
          s := r + 1
        ELSE
          INC(s);
        END
      END;
      (* We got to the end, so presumably "cand[r]" is now well-separated: *)
      INC(r);
    END;
  END MergePairOverlaps;

PROCEDURE Overlap(
    READONLY a, b: T;
    minSteps: NAT;
    maxAdjust: NAT;
    VAR (*WORK*) ra, rb: REF PZMatch.T;   (* Work areas *)
  ): BOOL =
  BEGIN
    RETURN PackedMatchOverlap(a.pm, b.pm, a.seg, b.seg, minSteps, maxAdjust, ra,rb)
  END Overlap;

PROCEDURE ReallocMatch(VAR r: REF PZMatch.T; np: NAT) =
  (*
    Reallocates "r^" if needed to store at least "np" elements.
    May discard the old contents. *)
  BEGIN
    IF r = NIL THEN
      r := NEW(REF PZMatch.T, np)
    ELSIF NUMBER(r^) < np THEN
      r := NEW(REF PZMatch.T, MAX(2*NUMBER(r^), np))
    END
  END ReallocMatch;

PROCEDURE PackedMatchOverlap(
    a, b: REF PZMatch.PackedT;
    READONLY aSeg, bSeg: PZSegment.Pair;
    minSteps: NAT;
    maxAdjust: NAT;
    VAR (*WORK*) ra, rb: REF PZMatch.T;   (* Work area *)
  ): BOOL =
  VAR na, nb: NAT;
  BEGIN
    IF a = NIL THEN na := MAX(aSeg[0].ns, aSeg[1].ns) ELSE na := a.np END;
    ReallocMatch(ra, na);
    IF b = NIL THEN nb := MAX(bSeg[0].ns, bSeg[1].ns) ELSE nb := b.np END;
    ReallocMatch(rb, nb);
    minSteps := MIN(MIN(na, nb) - 1, minSteps);
    WITH
      am = SUBARRAY(ra^, 0, na),
      bm = SUBARRAY(rb^, 0, nb)
    DO
      IF a = NIL THEN
        PZMatch.DoMostPerfect(aSeg[0].ns, aSeg[1].ns, am)
      ELSE
        PZMatch.DoUnpack(a^, am)
      END;
      IF b = NIL THEN
        PZMatch.DoMostPerfect(bSeg[0].ns, bSeg[1].ns, bm)
      ELSE
        PZMatch.DoUnpack(b^, bm)
      END;
      RETURN PZMatch.Overlap(am, bm, aSeg, bSeg, maxAdjust) >= 2*minSteps
    END
  END PackedMatchOverlap;

PROCEDURE Join(READONLY a, b: T): T =
  (*
    Joins two candidates, assuming that they belong to the
    same curve pair and they overlap.
  *)
  VAR
    c: T;
    oldSteps, newSteps: NAT := 0;
  BEGIN
    (* Compute the union of the two candidates on each curve: *)
    FOR j := 0 TO 1 DO
      c.seg[j] := PZSegment.Join(a.seg[j], b.seg[j]);
      oldSteps := oldSteps + (a.seg[j].ns + b.seg[j].ns) - 2;
      newSteps := newSteps + c.seg[j].ns
    END;
    <* ASSERT newSteps <= oldSteps  *>
    WITH
      r = FLOAT(newSteps, LONG)/FLOAT(oldSteps, LONG)
    DO
      c.length := r*(a.length + b.length);
      c.mismatch := 0.0d0;
      c.matchedLength := 0.0d0;
      c.pm := NIL
    END;
    RETURN c
  END Join;

PROCEDURE FLR(x: LONG; w, p: NAT): TEXT =
  BEGIN
    RETURN Fmt.Pad(Fmt.LongReal(x, Fmt.Style.Fix, prec := p), w)
  END FLR;

<*UNUSED*>
PROCEDURE FI(x: NAT; w: NAT): TEXT =
  BEGIN
    RETURN Fmt.Pad(Fmt.Int(x), w)
  END FI;

BEGIN
END PZCandidate.
(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
