MODULE PZFourier;
(* Last edited on 1999-08-06 22:06:00 by hcgl *)

IMPORT Math; 
FROM PZTypes IMPORT LONG, LONGS, INT, NAT;

PROCEDURE FFT(VAR V: LONGS; VAR F: LONGS) = 
  BEGIN 
    WITH N = NUMBER(V) DO 
      IF N = 1 THEN  F[0] := V[0] 
      ELSE
        WITH H = N DIV 2 DO
          <* ASSERT N = H + H *>
          FOR i:=0 TO H-1 DO  
            F[i] := V[i*2];
            F[i+H] := V[i*2+1] 
          END;
          FFT(SUBARRAY(F,0,H), SUBARRAY(V,0,H));
          FFT(SUBARRAY(F,H,H), SUBARRAY(V,H,H)); 
          FixFFT(V,F) 
        END
      END (* if *)
    END
  END FFT; 


PROCEDURE FixFFT(READONLY v: LONGS ; VAR F: LONGS) =
  VAR c2,s2,cj,sj : LONG; 
  CONST Pi = 3.14159265358979323d0;
  BEGIN
    WITH
      N =  NUMBER(v),
      H =  N DIV 2,
      c1 = Math.cos(2.0d0 * Pi / FLOAT(N,LONG)),
      s1 = Math.sin(2.0d0 * Pi / FLOAT(N,LONG)),
      coef = 1.0d0 / Math.sqrt(2.00d0)
    DO
      cj :=  c1; sj := s1;
      
      F[0] := coef * ( v[0] + v[H]);
      F[H] := coef * ( v[0] - v[H]); 
      
      FOR j:= 1 TO H-1 DO 
        F[j] := coef * ( v[j] + cj * v[j+H] - sj * v[N-j]);
        c2 := cj * c1 - sj * s1;
        s2 := sj * c1 + cj * s1;
        cj := c2; sj := s2;
      END;
      
      cj :=  c1; sj := s1;  
      
      FOR j:= H+1 TO N-1 DO 
        F[j] := coef * ( v[j-H] - cj * v[j] + sj * v[N+H-j]);
        c2 := cj * c1 - sj * s1;
        s2 := sj * c1 + cj * s1;
        cj := c2; sj := s2;
      END
    END
  END FixFFT;
  

PROCEDURE Order(band, step, length: LONG): NAT = 
  BEGIN
    RETURN PowGreaterThanN(CEILING(length/MIN(band/4.0d0, step))) 
  END Order;

  
PROCEDURE PowGreaterThanN(n: INT): INT =
  VAR x: NAT := 1;
  BEGIN 
    WHILE x <= n DO x := x + x END;
    RETURN x
  END PowGreaterThanN; 
  
  
PROCEDURE Diff(VAR F: T; order: NAT) =
  BEGIN
    WITH  
      N = NUMBER(F[0]),
      maxFreq = N DIV 2
    DO  
      IF order > 0 THEN 
        FOR k := 0 TO 2 DO F[k][0] := 0.0d0 END;
        FOR i := 1  TO maxFreq DO 
          FOR k := 0 TO 2 DO 
            WITH
              p = Math.pow(FLOAT(i,LONG),FLOAT(order, LONG)),
              u = p * F[k][i],
              v = p * F[k][N-i]
            DO
              IF k < maxFreq THEN 
                CASE order MOD 4 OF 
                  | 0 => F[k][i] := +u; F[k][N-i] := +v;
                  | 1 => F[k][i] := -v; F[k][N-i] := +u;
                  | 2 => F[k][i] := -u; F[k][N-i] := -v;
                  | 3 => F[k][i] := +v; F[k][N-i] := -u;
                ELSE  <* ASSERT FALSE *>    
                END
              ELSE
                CASE order MOD 4 OF 
                  | 0 => F[k][i] := +u;
                  | 1 => F[k][i] := 0.0d0
                  | 2 => F[k][i] := -u;
                  | 3 => F[k][i] := 0.0d0
                ELSE  <* ASSERT FALSE *>    
                END
              END
            END;
          END;  
        END;
      END;
    END;
  END Diff;
  
BEGIN   
END PZFourier.
(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
