MODULE PZFullPlot;

(* Last edited on 1999-09-04 19:47:25 by hcgl *)

IMPORT Text;
IMPORT PZPlot, PZLR3Chain, PZCandidate, PSPlot;

FROM PSPlot IMPORT Color; 
FROM PZTypes IMPORT LONG, NAT, BOOL;

CONST
  Aspect = 2.0d0;  (* Width/height of plotting area *)

PROCEDURE Candidate(
    fname: TEXT ;
    READONLY cand: PZCandidate.T;
    READONLY c0, c1: PZLR3Chain.T;  (* Chains, already mapped; *)
    whole: BOOL;
    closed: BOOL;
    dots: BOOL;
    pairing: BOOL;
    labels: BOOL;
    pointers: BOOL;
    axes: BOOL;
    epsFormat: BOOL;
    caption: TEXT := "";
    step: NAT := 1; 
  ) =
  CONST 
      PlotXSize = 150.0d0;
      PlotYSize = PlotXSize/Aspect;
  BEGIN
    IF epsFormat THEN
      WITH
        f = NEW(PSPlot.EPSFile).open(fname & ".eps", PlotXSize, PlotYSize)
      DO
        ScalesAndAxes(f, c0, c1, axes := axes);
        PZPlot.Candidate(
          f, cand, c0, c1,
          whole := whole, closed := closed, 
          dots := dots, pairing := pairing, labels := labels,
          pointers := pointers,
          step := step
        );
        f.close()     
      END
    ELSE    
      WITH
        f = NEW(PSPlot.PSFile).open(fname & ".ps")
      DO
        f.beginPage();
          PSBeginDrawing(f, PlotXSize, PlotYSize);
            ScalesAndAxes(f, c0, c1, axes := axes);
            PZPlot.Candidate(
              f, cand, c0, c1,
              whole := whole, closed := closed, 
              dots := dots, pairing := pairing, labels := labels,
              pointers := pointers,
              step := step
            );
            IF NOT Text.Empty(caption) THEN
              f.setTextColor(Color{0.0, 0.0, 0.0});
              f.caption(caption); 
            END;
          f.endDrawing();
        f.endPage();
        f.close()     
      END
    END
  END Candidate;
  

PROCEDURE PSBeginDrawing(f: PSPlot.PSFile; xSize, ySize: LONG) =
  CONST 
  BEGIN
    WITH
      xMargin = 0.5d0 * (PSPlot.LetterXSize - xSize),
      yMargin = 0.5d0 * (PSPlot.LetterYSize - ySize),
      xCenter = xSize * 0.5d0 + xMargin,
      yCenter = ySize * 0.5d0 + yMargin
    DO
      f.beginDrawing(xSize, ySize, xCenter, yCenter);
    END
  END PSBeginDrawing;


PROCEDURE ScalesAndAxes(
    f: PSPlot.File;
    READONLY c0, c1: PZLR3Chain.T;
    axes: BOOL;
  ) = 
  BEGIN
    WITH 
      b = BBoxExpand(BBoxJoin(BBoxGet(c0), BBoxGet(c1)), 0.05d0),
      xm = (b[0].lo + b[0].hi)/2.0d0,
      ym = (b[1].lo + b[1].hi)/2.0d0,
      hwx = (b[0].hi - b[0].lo)/2.0d0,
      hwy = (b[1].hi - b[1].lo)/2.0d0,
      fragWidth = MAX(hwx, Aspect*hwy),
      xMin = xm-fragWidth,
      xMax = xm+fragWidth,
      yMin = ym-fragWidth/Aspect,
      yMax = ym+fragWidth/Aspect
      (* , gridSpacing = SelectGridSpacing(hwx) *)
    DO 
      f.setScale(PSPlot.Axis.X, xMin, xMax);
      f.setScale(PSPlot.Axis.Y, yMin, yMax);
      IF axes THEN
        f.setLineWidth(0.2);
        f.setLineSolid();
        f.setLineColor(Color{0.75, 0.75, 0.75});
        f.coordLine(PSPlot.Axis.X, xm);
        f.coordLine(PSPlot.Axis.Y, ym);
      END;
      (* DrawGridLines(f, PSPlot.Axis.X, xMin, xMax, gridSpacing) *)
      (* DrawGridLines(f, PSPlot.Axis.Y, yMin, yMax, gridSpacing) *)
    END;
  END ScalesAndAxes;

TYPE
  Range = RECORD lo, hi: LONG END;
  BoundingBox = ARRAY [0..2] OF Range;

PROCEDURE BBoxGet(READONLY c: PZLR3Chain.T): BoundingBox =
  VAR b: BoundingBox;
  BEGIN
    FOR k := 0 TO 2 DO 
      WITH lo = b[k].lo, hi = b[k].hi DO
        lo := LAST(LONG);
        hi := FIRST(LONG);
        FOR i := 0 TO LAST(c) DO 
          lo := MIN(lo, c[i][k]);
          hi := MAX(hi, c[i][k])
        END;
      END
    END;
    RETURN b
  END BBoxGet;
  
PROCEDURE BBoxJoin(READONLY b1, b2: BoundingBox): BoundingBox =
  VAR b: BoundingBox;
  BEGIN
    FOR k := 0 TO 2 DO 
      b[k].lo := MIN(b1[k].lo, b2[k].lo);
      b[k].hi := MAX(b1[k].hi, b2[k].hi)
    END;
    RETURN b
  END BBoxJoin;
  
PROCEDURE BBoxExpand(READONLY b: BoundingBox; rel: LONG): BoundingBox =
  VAR r: BoundingBox;
  BEGIN
    FOR k := 0 TO 2 DO 
      WITH extra = rel*MAX(1.0d-100, b[k].hi - b[k].lo) DO
        r[k].lo := b[k].lo - extra;
        r[k].hi := b[k].hi + extra
      END
    END;
    RETURN r
  END BBoxExpand;
  
BEGIN
END PZFullPlot.
(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
