MODULE PZHistogram;

(* Last edited on 1999-11-12 23:18:22 by hcgl *)

IMPORT Wr, Thread, Fmt, Math;
FROM PZTypes IMPORT LONG, LONGS, NAT;

<* FATAL Wr.Failure, Thread.Alerted *>

PROCEDURE New(lo, hi: LONG; m: NAT): REF T =
  BEGIN
    <* ASSERT m >= 2 *>
    <* ASSERT hi > lo *>
    WITH 
      r = NEW(REF T),
      h = r^
    DO
      h := T{
        lo := lo,
        hi := hi,
        m := m,
        c := NEW(REF LONGS, m),
        wSum := 0.0d0,
        wxSum := 0.0d0,
        wxxSum := 0.0d0
      };
      FOR i := 0 TO LAST(h.c^) DO h.c^[i] := 0.0d0 END;
      RETURN r
    END
  END New;

PROCEDURE Add(VAR h: T; x: LONG; w: LONG := 1.0d0) =
  BEGIN
    WITH
      xx = MIN(h.hi, MAX(h.lo, x)),
      k = ROUND(FLOAT(h.m-1, LONG)*(xx - h.lo)/(h.hi - h.lo)),
      bin = h.c^[k]
    DO
      bin := bin + w;
      h.wSum := h.wSum + w;
      h.wxSum := h.wxSum + w*x;
      h.wxxSum := h.wxxSum + w*x*x
    END
  END Add;

PROCEDURE Output(wr: Wr.T; READONLY h: T) =
  BEGIN
    WITH
      m = h.m,
      c = h.c^,
      step = (h.hi - h.lo)/FLOAT(m-1, LONG)
    DO
      FOR i := 0 TO m-1 DO
        WITH
          fi = FLOAT(i, LONG),
          xLo = h.lo + step * (fi - 0.5d0),
          xMd = h.lo + step * fi,
          xHi = h.lo + step * (fi + 0.5d0)
        DO
          Wr.PutText(wr, " ");
          Wr.PutText(wr, FLR(xLo, 13,6));
          Wr.PutText(wr, " ");
          Wr.PutText(wr, FLR(xMd, 13,6));
          Wr.PutText(wr, " ");
          Wr.PutText(wr, FLR(xHi, 13,6));
          Wr.PutText(wr, " ");
          Wr.PutText(wr, FLR(c[i], 13,6));
          Wr.PutText(wr, "\n");
        END
      END;
    END
  END Output;

PROCEDURE Avg(READONLY h: T): LONG =
  BEGIN
    IF h.wSum = 0.0d0 THEN
      RETURN 0.0d0
    ELSE 
      RETURN h.wxSum/h.wSum
    END
  END Avg;
  
PROCEDURE Var(READONLY h: T): LONG =
  BEGIN
    IF h.wSum = 0.0d0 THEN
      RETURN 0.0d0
    ELSE
      RETURN (h.wxxSum - h.wxSum*h.wxSum/h.wSum)/h.wSum
    END
  END Var;
  
PROCEDURE Dev(READONLY h: T): LONG =
  BEGIN
    RETURN Math.sqrt(Var(h))
  END Dev;

PROCEDURE FLR(x: LONG; w, p: NAT): TEXT =
  BEGIN
    RETURN Fmt.Pad(Fmt.LongReal(x, Fmt.Style.Fix, prec := p), w)
  END FLR;

BEGIN
END PZHistogram.
(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
