MODULE PZIntChain;

(* Last edited on 1999-11-15 00:05:15 by hcgl *)

IMPORT  FileFmt, NGet, FGet, NPut, FPut, Rd, Wr, Thread;
IMPORT PZSegment;
FROM PZTypes IMPORT NAT;

PROCEDURE Trim(READONLY c: T; start, length: NAT): REF T =
  BEGIN
    WITH rx = NEW(REF T, length), x = rx^ DO
      DoTrim(c, start, length, x);
      RETURN rx
    END;
  END Trim;

PROCEDURE DoTrim(READONLY c: T; start, length: NAT; VAR x: T) =
  BEGIN
    WITH
      N = NUMBER(c),
      ini = start MOD N,
      fin = ini + length - 1
    DO
      IF fin < N  THEN
        x := SUBARRAY(c, ini, length)
      ELSE
        SUBARRAY(x, 0, N-ini) := SUBARRAY(c, ini, N-ini);
        SUBARRAY(x, N-ini, (fin+1) MOD N):= SUBARRAY(c, 0, (fin+1) MOD N);
      END;
    END;
  END DoTrim;

PROCEDURE Reverse(VAR c: T) =
  VAR aux: Point;
  BEGIN
    WITH U = LAST(c) DO
      FOR i:=0 TO U DIV 2  DO
        aux := c[i]; c[i] := c[U-i]; c[U-i] := aux
      END
    END
  END Reverse;

PROCEDURE ExtractSegment(READONLY c: T; s: PZSegment.T): REF T =
  BEGIN
    <* ASSERT NUMBER(c) = s.tot *>
    WITH rx = NEW(REF T, s.ns), x = rx^ DO
      DoExtractSegment(c, s, x);
      RETURN rx
    END
  END ExtractSegment;

PROCEDURE DoExtractSegment(READONLY c: T; s: PZSegment.T; VAR x: T) =
  BEGIN
    <* ASSERT NUMBER(c) = s.tot *>
    DoTrim(c, s.ini, s.ns, x);
    IF s.rev THEN Reverse(x) END;
  END DoExtractSegment;

CONST FileVersion = "97-01-21";

PROCEDURE Read(rd: Rd.T): ReadData =
  VAR
    data : ReadData;
  BEGIN
    FileFmt.ReadHeader(rd, "PZIntChain.T", FileVersion);
    data.cmt := FileFmt.ReadComment(rd, '|');
    WITH
      N = NGet.Int(rd, "points")
    DO
      FGet.EOL(rd);
      data.c := NEW(REF T, N);
      WITH c = data.c^ DO
        FOR i := 0 TO N-1  DO
          c[i,0] :=   FGet.Int(rd);
          c[i,1] :=   FGet.Int(rd);
        END;
        FileFmt.ReadFooter(rd, "PZIntChain.T");
        RETURN data
      END;
    END;
  END Read;

PROCEDURE Write(wr: Wr.T; cmt: TEXT; READONLY c: T) =
  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN
    FileFmt.WriteHeader(wr, "PZIntChain.T", FileVersion);
    FileFmt.WriteComment(wr, cmt, '|');
    NPut.Int(wr, "length", NUMBER(c));  FPut.EOL(wr);
    FOR i := 0 TO LAST(c)  DO
       FPut.Int(wr, c[i,0]);  FPut.Space(wr);
       FPut.Int(wr, c[i,1]);  FPut.EOL(wr)
    END;
    FileFmt.WriteFooter(wr, "PZIntChain.T");
    Wr.Flush(wr);
  END Write;

BEGIN
END PZIntChain.
(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
