MODULE PZLR3Chain;

(* Last edited on 1999-12-22 16:33:17 by hcgl *)

IMPORT NGet, NPut, FGet, FPut, FileFmt, FileRd, OSError, Fmt;
IMPORT Wr, Rd, LR3, LR4, LR4x4, Thread;
IMPORT PZMatch, PZProc, PZSegment, PZIntChain, PZLRChain, PZFourier, PZGeo;
FROM Math IMPORT sqrt;
FROM Stdio IMPORT stderr;
FROM PZTypes IMPORT LONG, LONGS, INT, NAT, BOOL, BOOLS;

PROCEDURE FromIntChain(READONLY c: PZIntChain.T): REF T =
  BEGIN
    WITH
      nc = NUMBER(c),
      rch = NEW(REF T, nc),
      ch = rch^
    DO
      FOR i := 0 TO nc-1 DO
        ch[i,0] := FLOAT(c[i,0], LONG);
        ch[i,1] := FLOAT(c[i,1], LONG);
        ch[i,2] := 0.0d00
      END;
      RETURN rch
    END
  END FromIntChain;

PROCEDURE ToIntChain(READONLY f: T): REF PZIntChain.T =
  BEGIN
    WITH
      nc = NUMBER(f),
      rc = NEW(REF PZIntChain.T, nc),
      c = rc^
    DO
      FOR i := 0 TO nc-1 DO
        c[i,0] := ROUND(f[i,0]);
        c[i,1] := ROUND(f[i,1])
      END;
      RETURN rc
    END
  END ToIntChain;

PROCEDURE FromLineSegment(READONLY p, q: LR3.T; n: NAT): REF T =
  BEGIN
    WITH
      delta = LR3.Scale(1.0d0/FLOAT(n-1, LONG), LR3.Sub(q, p)),
      rc = NEW(REF T, n),
      c = rc^
    DO
      FOR i:= 0 TO n-1 DO
        WITH f = FLOAT(i,LONG), ci = c[i] DO
          ci[0] := delta[0] * f + p[0];
          ci[1] := delta[1] * f + p[1];
          ci[2] := delta[2] * f + p[2];
        END
      END;
      RETURN rc;
    END
  END FromLineSegment;

PROCEDURE Trim(READONLY c: T; start, length: NAT): REF T =
  BEGIN
    WITH rx = NEW(REF T, length), x = rx^ DO
      DoTrim(c, start, length, x);
      RETURN rx
    END;
  END Trim;

PROCEDURE DoTrim(READONLY c: T; start, length: NAT; VAR x: T) =
  BEGIN
    WITH
      nc = NUMBER(c),
      ini = start MOD nc,
      fin = ini + length - 1
    DO
      IF fin < nc  THEN
        x := SUBARRAY(c, ini, length)
      ELSE
        SUBARRAY(x, 0, nc-ini) := SUBARRAY(c, ini, nc-ini);
        SUBARRAY(x, nc-ini, (fin+1) MOD nc):= SUBARRAY(c, 0, (fin+1) MOD nc);
      END;
    END;
  END DoTrim;

PROCEDURE Reverse(VAR c: T) =
  VAR aux: LR3.T;
  BEGIN
    WITH U = LAST(c) DO
      FOR i:=0 TO U DIV 2  DO
        aux := c[i]; c[i] := c[U-i]; c[U-i] := aux
      END
    END
  END Reverse;

PROCEDURE ExtractSegment(READONLY c: T; s: PZSegment.T): REF T =
  BEGIN
    <* ASSERT NUMBER(c) = s.tot *>
    WITH rx = NEW(REF T, s.ns), x = rx^ DO
      DoExtractSegment(c, s, x);
      RETURN rx
    END
  END ExtractSegment;

PROCEDURE DoExtractSegment(READONLY c: T; s: PZSegment.T; VAR x: T) =
  BEGIN
    <* ASSERT NUMBER(c) = s.tot *>
    DoTrim(c, s.ini, s.ns, x);
    IF s.rev THEN Reverse(x) END;
  END DoExtractSegment;

PROCEDURE Map(READONLY c: T; READONLY m: LR4x4.T): REF T =
  BEGIN
    WITH
      rd = NEW(REF T, NUMBER(c)),
      d = rd^
    DO
      d := c; 
      DoMap(d, m);
      RETURN rd
    END
  END Map;

PROCEDURE DoMap(VAR c: T; READONLY m: LR4x4.T) =
  BEGIN
    FOR k:= 0 TO LAST(c) DO
      WITH
        ck = c[k],
        u = LR4.T{1.0d0, ck[0], ck[1], ck[2]},
        v = LR4x4.MapRow(u, m)
      DO
        ck := LR3.T{v[1]/v[0], v[2]/v[0], v[3]/v[0]}
      END
    END
  END DoMap;

PROCEDURE AlignmentMatrix(READONLY a, b: T): LR4x4.T =
  BEGIN
    WITH 
      na = NUMBER(a),
      ac = (na - 1) DIV 2,
      ai = MAX(0, MIN(na DIV 5, ac - 1)), 
      af = MIN(na - 1, MAX(4*na DIV 5, ac + 1)),
      
      pi = a[ai],
      pf = a[af],
      pc = Barycenter(a),
      
      nb = NUMBER(b),
      bc = (nb - 1) DIV 2,
      bi = MAX(0, MIN(nb DIV 5, bc - 1)),
      bf = MIN(nb - 1, MAX(4*nb DIV 5, bc + 1)),
      
      qi = b[bi],
      qf = b[bf],
      qc = Barycenter(b),
      
      u  = PZGeo.SegDir(pi, pf),
      v  = PZGeo.SegDir(qi, qf),
      T1 = PZGeo.Translation(LR3.Neg(pc)),
      T2 = PZGeo.Translation(qc),
      R =  PZGeo.Rotation(u, v)
    DO
      RETURN LR4x4.Mul(LR4x4.Mul(T1, R), T2);
    END;
  END AlignmentMatrix;

PROCEDURE LinearLengths(
    READONLY p: T;              (* "p[i]" is curve position at time "t[i]". *)
    VAR s: PZLRChain.T;         (* "s[j]" is the length from "p[0] to "p[j] "j". *)
  ): LONG =
  VAR ns: LONG := 0.0d0;
  BEGIN
    WITH
      m = NUMBER(p)
    DO
      <* ASSERT NUMBER(s) = m *>
      s[0] := 0.0d0;
      FOR i := 1 TO LAST(p) DO
        ns := ns + LR3.Dist(p[i], p[i-1]);
        s[i] := ns;
      END;
      RETURN ns + LR3.Dist(p[m-1], p[0]);
    END
  END LinearLengths;

PROCEDURE LinearUniformSample(
    READONLY t: PZLRChain.T;
    tPeriod: LONG;
    READONLY p: T;
    tStart: LONG;
    VAR q: T;
  ) =
  VAR ib: NAT;
      ta, tb: LONG;
      k: INT;
  BEGIN
    WITH
      np = NUMBER(p),
      nq = NUMBER(q),
      tHi = t[0] + tPeriod
    DO
      <* ASSERT NUMBER(t) = np *>
      PZProc.ReduceToPeriod(tStart, t[0], tHi, k);
      ta := t[0]; ib := 1; tb := t[1];
      FOR j := 0 TO nq-1 DO
        WITH tau = tStart + tPeriod * (FLOAT(j, LONG) / FLOAT(nq, LONG)) DO
          <* ASSERT tau >= t[ib-1] *>
          <* ASSERT tau < tHi *>
          WHILE ib < np AND tau > t[ib] DO
            ta := tb; ib := ib+1;
            IF ib = np THEN tb := tHi ELSE tb := t[ib] END;
          END;
          q[j] := PZGeo.LinearInterpolate(tau, ta, p[ib-1], tb, p[ib MOD np])
        END
      END
    END
  END LinearUniformSample;

PROCEDURE LinearEquidistantSample(READONLY p: T; tStart: LONG; VAR q: T) =
  BEGIN
    WITH
      np = NUMBER(p),
      t = NEW(REF PZLRChain.T, np)^,
      length = LinearLengths(p, t)
    DO
      LinearUniformSample(t, length, p, tStart*length, q)
    END
  END LinearEquidistantSample;

PROCEDURE LinearTimeSample(
    READONLY t: PZLRChain.T;
    tPeriod: LONG;
    READONLY p: T;
    READONLY r: PZLRChain.T;
    VAR q: T;
  ) =
  VAR ia, ib: NAT;
      k: INT;
      ta, tb, tau: LONG;
  BEGIN
    WITH
      np = NUMBER(p),
      nq = NUMBER(q),
      tLo = t[0],
      tHi = t[0] + tPeriod
    DO
      <* ASSERT NUMBER(t) = np *>
      <* ASSERT NUMBER(r) = nq *>
      ia := 0; ta := t[0];
      ib := 1; tb := t[1];
      FOR j:= 0 TO nq-1 DO
        (* Invariant: "ia IN [0..np-1]", "ib = ia+1" *)
        tau := r[j];
        PZProc.ReduceToPeriod(tau, tLo, tHi, k);
        <* ASSERT tLo <= tau AND tau < tHi *>
        WHILE tau < t[ia] DO ib := ia; DEC(ia) END;
        WHILE ib < np AND tau > t[ib] DO ia := ib; INC(ib) END;
        WITH
          ta = t[ia],
          tb = t[ib MOD np] + tPeriod * FLOAT(ib DIV np, LONG)
        DO
          q[j] := PZGeo.LinearInterpolate(tau, ta, p[ia MOD np], tb, p[ib MOD np]);
        END;
      END;
    END
  END LinearTimeSample;

PROCEDURE EstimateVelocities(
    READONLY t: PZLRChain.T;
    tPeriod: LONG;
    READONLY p: T;
    VAR dp: T;
    est: VelEstimator;
  ) =
  VAR ka, kb: NAT;
      sa, sb: LONG := 0.0d0;
  BEGIN
    WITH
      np = NUMBER(p)
    DO
      <* ASSERT NUMBER(dp) = np *>
      <* ASSERT NUMBER(t) = np *>
      IF np <= 2 THEN
        FOR i := 0 TO np-1 DO dp[i] := LR3.T{0.0d0, 0.0d0, 0.0d0} END;
      ELSE
        (* Estimate the velocities by "L" method: *)
        ka := np-2; kb := np-1;
        sa := t[np-2] - tPeriod; sb := t[np-1] - tPeriod;
        FOR i:= 0 TO np-1 DO
          dp[kb] := est(sa, p[ka], sb, p[kb], t[i], p[i]);
          ka := kb; kb := i;
          sa := sb; sb := t[i];
        END
      END
    END
  END EstimateVelocities;

PROCEDURE HermiteUniformSample(
    READONLY t: PZLRChain.T;
    tPeriod: LONG;
    READONLY p: T;
    READONLY dp: T;
    tStart: LONG;
    VAR q: T;
    VAR dq: T;
  ) =
  VAR ib: NAT;
      ta, tb: LONG;
      k: INT;
  BEGIN
    WITH
      np = NUMBER(p),
      nq = NUMBER(q),
      tLo = t[0],
      tHi = tLo + tPeriod
    DO
      <* ASSERT NUMBER(dp) = np *>
      <* ASSERT NUMBER(t) = np *>
      PZProc.ReduceToPeriod(tStart, tLo, tHi, k);
      ta := t[0]; ib := 1; tb := t[1];
      FOR j := 0 TO nq-1 DO
        WITH tau = tStart + tPeriod * (FLOAT(j, LONG) / FLOAT(nq, LONG)) DO
          <* ASSERT tau >= t[ib-1] *>
          <* ASSERT tau < tHi *>
          WHILE ib < np AND tau > t[ib] DO
            ta := tb; ib := ib + 1;
            IF ib = np THEN tb := tHi ELSE tb := t[ib] END;
          END;
          <* ASSERT tau >= ta AND tau <= tb *>
          PZGeo.HermiteInterpolate(
            tau,
            ta,
            p[ib-1], dp[ib-1],
            tb,
            p[ib MOD np], dp[ib MOD np],
            q[j], dq[j]
          )
        END
      END;
    END
  END HermiteUniformSample;

PROCEDURE HermiteEquidistantSample(
    READONLY p: T;
    tStart: LONG;
    VAR q: T;
    VAR dq: T;
  ) =
  BEGIN
    IF NUMBER(p) <= 1 THEN
      FOR i := 0 TO LAST(q) DO
        q[i] := p[0]; dq[i] := LR3.T{0.0d0, 0.0d0, 0.0d0}
      END;
      RETURN
    END;
    WITH
      np = NUMBER(p),
      t = NEW(REF PZLRChain.T, np)^,
      length = LinearLengths(p, t),
      dp = NEW(REF T, np)^
    DO
      EstimateVelocities(t, length, p, dp, PZGeo.EstimateVelocityL);
      HermiteUniformSample(t, length, p, dp, tStart*length, q, dq);
    END
  END HermiteEquidistantSample;

PROCEDURE HermiteTimeSample(
    READONLY t: PZLRChain.T;
    tPeriod: LONG;
    READONLY p: T;
    READONLY dp: T;
    READONLY r: PZLRChain.T;
    VAR q: T;
    VAR dq: T;
  ) =
  VAR ia, ib: NAT;
      k: INT;
      tau: LONG;
  BEGIN
    WITH
      np = NUMBER(p),
      nq = NUMBER(q),
      tLo = t[0],
      tHi = tLo + tPeriod
    DO
      <* ASSERT NUMBER(t) = np *>
      <* ASSERT NUMBER(r) >= nq *>
      ia := 0; ib := 1;
      FOR j := 0 TO nq-1 DO
        (* Invariant: "ia IN [0..np-1]", "ib = ia+1" *)
        tau := r[j];
        PZProc.ReduceToPeriod(tau, tLo, tHi, k);
        WHILE tau < t[ia] DO ib := ia; DEC(ia) END;
        WHILE ib < np AND tau > t[ib] DO ia := ib; INC(ib) END;
        WITH
          ta = t[ia],
          tb = t[ib MOD np] + tPeriod * FLOAT(ib DIV np, LONG)
        DO
          <* ASSERT ta <= tau AND tau <= tb *>
          PZGeo.HermiteInterpolate(
            tau,
            ta, p[ia],           dp[ia],
            tb, p[(ia+1) MOD np], dp[(ia+1) MOD np],
            q[j], dq[j]
          )
        END;
      END;
    END
  END HermiteTimeSample;

PROCEDURE Barycenter(READONLY c: T): LR3.T =
  VAR p: LR3.T := LR3.T{0.0d0, ..};
  BEGIN
    FOR i := 0 TO LAST(c) DO
      WITH ci = c[i] DO
        p[0] := p[0] + ci[0];
        p[1] := p[1] + ci[1];
        p[2] := p[2] + ci[2];
      END
    END;
    WITH fn = FLOAT(NUMBER(c), LONG) DO
      p[0] := p[0] /fn;
      p[1] := p[1] /fn;
      p[2] := p[2] /fn;
    END;
    RETURN p
  END Barycenter;

CONST FileVersion = "97-10-29";

PROCEDURE Read(rd: Rd.T; headerOnly: BOOL): ReadData =

  PROCEDURE ToLong(n: INT): LONG =
    BEGIN
      <* ASSERT n > FIRST(INT) *>
      <* ASSERT n < LAST(INT) *>
      RETURN FLOAT(n, LONG)
    END ToLong;

  VAR d : ReadData;
  BEGIN
    FileFmt.ReadHeader(rd, "PZLR3Chain.T", FileVersion);
    d.cmt := FileFmt.ReadComment(rd, '|');
    d.unit := NGet.LongReal(rd, "unit");
    FGet.Skip(rd);
    d.samples := NGet.Int(rd, "samples"); FGet.EOL(rd);
    WITH
      N = d.samples,
      unit = d.unit
    DO
      IF headerOnly THEN
        d.c := NIL
      ELSE
        d.c := NEW(REF T, N);
        WITH c = d.c^ DO
          FOR i := 0 TO N-1 DO
            WITH ci = c[i] DO
              FGet.Skip(rd); ci[0] := ToLong(FGet.Int(rd)) * unit;
              FGet.Skip(rd); ci[1] := ToLong(FGet.Int(rd)) * unit;
              FGet.Skip(rd); ci[2] := ToLong(FGet.Int(rd)) * unit;
            END
          END
        END;
        FileFmt.ReadFooter(rd, "PZLR3Chain.T")
      END;
      RETURN d
    END
  END Read;

PROCEDURE Write(wr :Wr.T; READONLY cmt : TEXT;  READONLY c: T; unit: LONG) =

  PROCEDURE ToInt(x: LONG): INT =
    BEGIN
      <* ASSERT x > FLOAT(FIRST(INT),LONG) *>
      <* ASSERT x < FLOAT(LAST(INT), LONG) *>
      RETURN ROUND(x)
    END ToInt;

  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN
    FileFmt.WriteHeader(wr, "PZLR3Chain.T", FileVersion);
    FileFmt.WriteComment(wr, cmt, '|');
    NPut.LongReal(wr, "unit", unit); FPut.EOL(wr);
    NPut.Int(wr, "samples", NUMBER(c));  FPut.EOL(wr);
    FOR i := 0 TO LAST(c)  DO
      WITH ci = c[i] DO
        FPut.Int(wr, ToInt(ci[0] / unit));  FPut.Space(wr);
        FPut.Int(wr, ToInt(ci[1] / unit));  FPut.Space(wr);
        FPut.Int(wr, ToInt(ci[2] / unit));  FPut.EOL(wr)
      END
    END;
    FileFmt.WriteFooter(wr, "PZLR3Chain.T");
    Wr.Flush(wr)
  END Write;

PROCEDURE ReadAll(
    prefix: TEXT;
    band: NAT;
    extension: TEXT;
    READONLY sel: BOOLS;
    headerOnly: BOOL;
    dir: TEXT := ".";
  ): ReadAllData =
  VAR unitMin: LONG := LAST(LONG);
      unitMax: LONG := FIRST(LONG);
      cmt: TEXT := ""; 
  CONST NoData = ReadData{ cmt := "NOT READ", samples := 0, c := NIL, unit := 0.0d0};
  <* FATAL Wr.Failure, Thread.Alerted, Rd.Failure *>
  BEGIN
    Wr.PutText(stderr, "reading geometric chains:\n");
    WITH
      rr = NEW(REF ARRAY OF ReadData, NUMBER(sel)), r = rr^
    DO
      FOR k := 0 TO LAST(sel) DO
        IF sel[k] THEN
          Wr.PutText(stderr, "  " & Fmt.Pad(Fmt.Int(k), 4, '0') & ": ");
          WITH
            chainTag  = Fmt.Pad(Fmt.Int(k), 4, '0'),
            bandTag = Fmt.Pad(Fmt.Int(band), 3, '0'),
            fileName = prefix & bandTag & extension,
            fullName = dir & "/" & chainTag &"/"& fileName,
            rd = OpenRead(fullName)
          DO
            IF rd # NIL THEN
              WITH data = Read(rd, headerOnly := headerOnly) DO
                Wr.PutText(stderr, "read from " & fullName & "\n");
                r[k] := data;
                <* ASSERT data.unit > 0.0d0 *>
                IF unitMin > unitMax THEN 
                  cmt := data.cmt
                END;
                unitMin := MIN(unitMin, data.unit);
                unitMax := MAX(unitMax, data.unit);
              END;
              Rd.Close(rd)
            ELSE
              r[k] := NoData;
              Wr.PutText(stderr, "file " & fullName & " not found.\n");
            END
          END
        ELSE
          r[k] := NoData;
        END
      END;
      RETURN ReadAllData {
        chain := rr,
        unitMin := unitMin,
        unitMax := unitMax,
        cmt := 
          "Chains read from " & 
          dir & "/NNNN/" & prefix & "BBB" & extension & "\n" &
          cmt
      }
    END
  END ReadAll;

PROCEDURE OpenRead(name: TEXT): Rd.T =
  BEGIN
    TRY
      RETURN FileRd.Open(name)
    EXCEPT
      OSError.E => RETURN NIL
    END
  END OpenRead;

PROCEDURE FourierTransform(READONLY  c: T; VAR f: Spectrum) =
  BEGIN
    WITH
      nc = NUMBER(c),
      r = NEW(REF PZLRChain.T, nc)^
    DO
      FOR i := 0 TO 2 DO
        FOR j := 0 TO nc-1 DO
          r[j] := c[j,i]
        END;
        PZFourier.FFT(r,f[i])
      END;
    END
  END FourierTransform;

PROCEDURE InvFourierTransform(VAR f: Spectrum; VAR c: T) =
  VAR
    r : REF PZLRChain.T;
  BEGIN
    WITH
      nc = NUMBER(c)
    DO
      r:= NEW(REF PZLRChain.T, nc);
      FOR i := 0 TO 2 DO
        PZFourier.FFT(f[i], r^);
        FOR j := 0 TO nc-1 DO
          c[j,i] := r[j];
        END
      END
    END
  END InvFourierTransform;

PROCEDURE FindBestMatch(
    READONLY a: T;
    READONLY b : T;
    maxDistSqr: LONG;
    removeUnmatchedEnds: BOOL;
    VAR (*OUT*) mt: REF PZMatch.T;
    VAR (*OUT*) avgDist: LONG;
    VAR (*OUT*) length: LONG;
    VAR (*OUT*) matchedLength: LONG;
    VAR (*WORK*) rCost: REF PZMatch.CostMatrix;
  )  =

  TYPE Pair = PZMatch.Pair;

  PROCEDURE StepDistSqr (READONLY p, q: Pair): LONG =
    (*
      The mean square distance between the chain steps
      "a[p[0]]->a[q[0]]" and "b[p[1]]->b[q[1]]". *)
    BEGIN
      WITH
        a1 = a[p[0]], b1 = b[p[1]],
        a2 = a[q[0]], b2 = b[q[1]],
        mdSqr = 0.5d0 * (LR3.DistSqr(a1,b1) + LR3.DistSqr(a2,b2))
      DO
        RETURN MIN(mdSqr, maxDistSqr)
      END
    END StepDistSqr;

  PROCEDURE StepLength (READONLY p, q: Pair): LONG =
    (*
      The length of the step "p->q", i.e. the mean length of
      "a[p[0]]--a[q[0]]" and "b[p[1]]--b[q[1]]". *)
    BEGIN
      WITH
        a1 = a[p[0]],
        a2 = a[q[0]],
        b1 = b[p[1]],
        b2 = b[q[1]],
        La = LR3.Dist(a1,a2),
        Lb = LR3.Dist(b1,b2)
      DO
        RETURN  (La + Lb) * 0.5d0;
      END
    END StepLength;

  PROCEDURE StepIntgDistSqr (READONLY p, q: Pair): LONG =
    (*
      A "PZMatch.StepFunction" whose value is the integrated square distance
      of the step "p -> q", i.e. between the chain steps
      "a[p[0]]->a[q[0]]" and "b[p[1]]->b[q[1]]". *)
    BEGIN
      RETURN StepDistSqr(p, q) * StepLength(p, q)
    END StepIntgDistSqr;

  PROCEDURE StepIsMatched(READONLY p, q: Pair): BOOL =
    (*
      A "PZMatch.StepPredicate" that returns TRUE if the step "p->q"
      has average chain distance less than "maxDist". *)
    BEGIN
      RETURN StepDistSqr(p, q) < maxDistSqr
    END StepIsMatched;

  VAR intgDistSqr: LONG;
  BEGIN
    WITH
      NA = NUMBER(a),
      NB = NUMBER(b)
    DO
      PZMatch.FindBest(
        NA, NB,
        stepCostFn := StepIntgDistSqr,
        (*OUT*) mt := mt, totCost := intgDistSqr,
        (*WORK*) rCost := rCost
      );
      IF removeUnmatchedEnds THEN
        mt := PZMatch.RemoveUnmatchedEnds(mt^, StepIsMatched);
        intgDistSqr := PZMatch.Integral(mt^,
          lengthFn := StepLength,
          stepFn := StepDistSqr
        )
      END;
      <* ASSERT NUMBER(mt^) > 0 *>
      length := PZMatch.Length(mt^, lengthFn := StepLength);
      IF length <= 0.0d0 THEN
        WITH cp = NUMBER(mt^) DIV 2, ctr = mt[cp] DO
          avgDist := LR3.DistSqr(a[ctr[0]], b[ctr[1]])
        END
      ELSE
        avgDist := sqrt(intgDistSqr/length)
      END;
      matchedLength := PZMatch.MatchedLength(mt^,
        lengthFn := StepLength,
        isMatched := StepIsMatched
      );
    END
  END FindBestMatch;

PROCEDURE RefineMatch(
    READONLY a: T;
    READONLY b: T;
    step: LONG;
    READONLY mtOld: PZMatch.T;
    maxAdjust: NAT;
    critDistSqr: LONG;
    maxDistSqr: LONG;
    VAR (*OUT*) mt: REF PZMatch.T;
    VAR (*OUT*) mismatch: LONG;
    VAR (*OUT*) length: LONG;
    VAR (*OUT*) matchedLength: LONG;
    VAR (*OUT*) minAvgDist: LONGS;
    VAR (*WORK*) rCost: REF PZMatch.CostMatrix;
  ) =

  TYPE Pair = PZMatch.Pair;

  PROCEDURE StepDistSqr (READONLY p, q: Pair): LONG =
    (*
      The mean square distance between the chain steps
      "a[p[0]]->a[q[0]]" and "b[p[1]]->b[q[1]]". *)
    BEGIN
      WITH
        a1 = a[p[0]],
        a2 = a[q[0]],
        b1 = b[p[1]],
        b2 = b[q[1]],
        mdSqr = 0.5d0 * (LR3.DistSqr(a1,b1) + LR3.DistSqr(a2,b2))
      DO
        RETURN MIN(mdSqr, maxDistSqr)
      END
    END StepDistSqr;

  PROCEDURE StepMismatch (READONLY p, q: Pair): LONG =
    (*
      A "PZMatch.StepFunction" whose value is the mismatch
      of the step "p -> q", i.e. the integrated square
      distance between the chain steps "a[p[0]]->a[q[0]]"
      and "b[p[1]]->b[q[1]]", minus the critical distance squared,
      all times the step length. *)
    BEGIN
      WITH d2 = StepDistSqr(p, q) - critDistSqr DO
        IF p[0] = q[0] OR p[1] = q[1] THEN
          RETURN step * d2 / 2.0d0
        ELSE
          RETURN step * d2
        END
      END;
    END StepMismatch;

  PROCEDURE StepIsMatched(READONLY p, q: Pair): BOOL =
    (*
      A "PZMatch.StepPredicate" that returns TRUE if the step "p->q"
      has average chain distance less than "maxDist". *)
    BEGIN
      RETURN StepDistSqr(p, q) < maxDistSqr
    END StepIsMatched;

  BEGIN
    WITH
      NA = NUMBER(a),
      NB = NUMBER(b)
    DO
      <* ASSERT NA > 0 *>
      <* ASSERT NB > 0 *>
      PZMatch.Refine(
        NA := NA, NB := NB,
        stepCostFn := StepMismatch,
        mtOld := mtOld,
        maxAdjust := maxAdjust,
        (*OUT*) mt := mt,
        (*OUT*) totCost := mismatch,
        (*OUT*) minTotCost := minAvgDist,
        (*WORK*) rCost := rCost
      );
      <* ASSERT  NUMBER(mt^) > 0 *>
      WITH
        mt = mt^,
        m = NUMBER(mt),
        stepCount = (mt[m-1][0] - mt[0][0]) + (mt[m-1][1] - mt[0][1]),
        matchedStepCount = PZMatch.MatchedStepCount(mt, StepIsMatched)
      DO
        length := FLOAT(stepCount, LONG) * step / 2.0d0;
        matchedLength := step * FLOAT(matchedStepCount, LONG)/2.0d0
      END;
      (* Compute the minimum average distance for each matching length "K". *)
      (* Note that "PZMatch.Refine" returns total mismatch, not just "intgDistSqr" *)
      <* ASSERT  minAvgDist[0] = 0.0d0 *>
      FOR K := 1 TO LAST(minAvgDist) DO
        WITH Lk = step*FLOAT(K, LONG)/2.0d0 DO
          minAvgDist[K] := sqrt(MAX(0.0d0, minAvgDist[K]/Lk + critDistSqr))
        END
      END
    END;
  END RefineMatch;

BEGIN
END PZLR3Chain.
(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
