MODULE PZSegment;

(* Last edited on 1999-11-16 20:10:29 by hcgl *)

IMPORT Rd, Wr, Fmt, FGet, NGet, FPut, NPut, FileFmt, Thread;

FROM PZTypes IMPORT LONG, INT, NAT, NATS, BOOL, BOOLS;

PROCEDURE IsEmpty(READONLY s: T): BOOL =
  BEGIN
    RETURN s.ns = 0
  END IsEmpty;

PROCEDURE AbsIndex(iRel: INT; READONLY s: T): INT =
  BEGIN
    IF s.rev THEN
      RETURN s.ini + s.ns - 1 - iRel
    ELSE
      RETURN s.ini + iRel
    END
  END AbsIndex;

PROCEDURE RelIndex(iAbs: INT; READONLY s: T): INT =
  BEGIN
    IF s.rev THEN
      RETURN s.ini + s.ns - 1 - iAbs
    ELSE
      RETURN iAbs - s.ini
    END
  END RelIndex;

PROCEDURE AbsIndexClip(iRel: NAT; READONLY s: T): NAT =
  BEGIN
    <* ASSERT iRel <= s.ns *>
    IF s.rev THEN
      RETURN (s.ini + s.ns - 1 - iRel) MOD s.tot
    ELSE
      RETURN (iRel + s.ini) MOD s.tot
    END
  END AbsIndexClip;

PROCEDURE RelIndexClip(iAbs: NAT; READONLY s: T): NAT =
  BEGIN
    WITH j = (iAbs - s.ini) MOD s.tot DO
      IF j < s.ns THEN
        IF s.rev THEN
          RETURN s.ns - 1 - j
        ELSE
          RETURN j
        END
      ELSE
        RETURN LAST(NAT)
      END
    END
  END RelIndexClip;

PROCEDURE Complement(READONLY s: T): T =
  BEGIN
    <* ASSERT s.ns > 0 *>
    <* ASSERT s.ns <= s.tot + 1 *>
    RETURN T {
      cvx := s.cvx,
      tot := s.tot,
      ini := (s.ini + s.ns - 1) MOD s.tot,
      ns := s.tot - s.ns + 2,
      rev := s.rev
    }
  END Complement;

PROCEDURE Expand(
    READONLY s: T;                 (* Segment to expand *)
    iniExtra, finExtra: INT;   (* Number of steps to add/delete. *)
  ): T =
  BEGIN
    WITH
      askedLen = s.ns + iniExtra + finExtra,
      newLen = MAX(1, MIN(askedLen, s.tot + 1)),
      excessSteps = askedLen - newLen
    DO
      RETURN T{
        cvx := s.cvx,
        tot := s.tot,
        ini := (s.ini - iniExtra +  excessSteps DIV 2) MOD s.tot,
        ns := newLen,
        rev := s.rev
      }
    END
  END Expand;

PROCEDURE Overlap(READONLY a, b: T): NAT =
  BEGIN
    IF a.cvx # b.cvx OR a.rev # b.rev THEN
      RETURN 0
    ELSE
      <* ASSERT a.tot = b.tot *>
      IF a.ns >= a.tot OR b.ns >= b.tot THEN
        RETURN MIN(a.ns, b.ns)
      ELSIF a.ini = b.ini THEN
        RETURN MIN(a.ns, b.ns)
      ELSE
        WITH
          m = a.tot,
          dab = (b.ini - a.ini) MOD m,
          nab = MIN(a.ns - dab, b.ns),
          dba = (a.ini - b.ini) MOD m,
          nba = MIN(b.ns - dba, a.ns)
        DO
          RETURN MAX(0, MAX(nba, nab))
        END
      END
    END
  END Overlap;

PROCEDURE Meet(READONLY a, b: T): T =
  VAR ini, ns: NAT;
  BEGIN
    IF a.cvx # b.cvx OR a.rev # b.rev THEN
      RETURN Empty
    ELSE
      <* ASSERT a.tot = b.tot *>
      IF a.ns = 0 OR b.ns = 0 THEN
        RETURN Empty
      ELSIF a.ns >= a.tot + 1 THEN
        RETURN b
      ELSIF b.ns >= b.tot + 1 THEN
        RETURN a
      END;
      WITH
        m = a.tot,
        dab = (b.ini - a.ini) MOD m,
        dba = (a.ini - b.ini) MOD m
      DO
        IF dab = 0 THEN
          (* Both segments start together *)
          ini := a.ini MOD m; ns := MIN(a.ns, b.ns)
        ELSIF dab < a.ns AND dba < b.ns THEN
          (* Each segment starts inside the other; wrap-around! *)
          RETURN Empty
        ELSIF dab < a.ns THEN
          (* "b" starts inside "a" *)
          ini := b.ini MOD m; ns := MIN(a.ns - dab, b.ns)
        ELSIF dba < b.ns THEN
          (* "a" starts inside "b" *)
          ini := a.ini MOD m; ns := MIN(a.ns, b.ns - dba)
        ELSE
          (* Disjoint *)
          RETURN Empty
        END;
        RETURN T{cvx := a.cvx, tot := m, ini := ini, ns := ns, rev := a.rev}
      END
    END
  END Meet;

PROCEDURE Join(READONLY a, b: T): T =
  VAR res: T;
  BEGIN
    <* ASSERT a.cvx = b.cvx *>
    <* ASSERT a.tot = b.tot *>
    <* ASSERT a.rev = b.rev *>
    res.cvx := a.cvx; res.tot := a.tot; res.rev := a.rev;
    WITH
      m = a.tot,
      dab = (b.ini - a.ini) MOD m,
      dba = m - dab,
      tab = dab < a.ns,  (* Segment "a" runs into segment "b" *)
      tba = dba < b.ns   (* Segment "b" runs into segment "a" *)
    DO
      IF (tab AND tba) OR a.ns > a.tot OR b.ns > b.tot THEN
        (* The two segments cover the whole contour: *)
        res.ini := 0; res.ns := a.tot+1;
      ELSE
        IF tab THEN
          (* Segment "a" runs over "b" but there is a gap from "b" to "a": *)
          res.ini := a.ini; res.ns := MAX(a.ns, dab + b.ns);
        ELSIF tba THEN
          (* Segment "b" runs over "a" but there is a gap from "a" to "b": *)
          res.ini := b.ini; res.ns := MAX(b.ns, dba + a.ns);
        ELSE
          (* The segments don't overlap; join through the smallest gap: *)
          IF dab - a.ns <= dba - b.ns THEN 
            res.ini := a.ini; res.ns := dab + b.ns
          ELSE
            res.ini := b.ini; res.ns := dba + a.ns
          END
        END;
        <* ASSERT res.ns <= m *>
      END
    END;
    RETURN res
  END Join;

CONST
  OldFileVersion = "97-04-05";
  NewFileVersion = "99-07-25";

PROCEDURE Write(
    wr :Wr.T;
    cmt: TEXT;
    READONLY s: List;
  ) =
  <* FATAL Thread.Alerted, Wr.Failure *>
  BEGIN
    WITH
      n = LAST(s)
    DO
      FileFmt.WriteHeader(wr, "PZSegment.List", NewFileVersion);
      FileFmt.WriteComment(wr, cmt, '|');
      NPut.Int(wr, "segments", NUMBER(s));        FPut.EOL(wr);
      FOR i := 0 TO n  DO
        WITH si = s[i] DO
          FPut.Int(wr, si.cvx); FPut.Space(wr);
          FPut.Int(wr, si.tot); FPut.Space(wr);
          FPut.Int(wr, si.ini); FPut.Space(wr);
          FPut.Int(wr, si.ns); FPut.Space(wr);
          IF si.rev THEN
            FPut.Char(wr, '-')
          ELSE
            FPut.Char(wr, '+')
          END;
          FPut.EOL(wr);
        END
      END;
      FileFmt.WriteFooter(wr, "PZSegment.List");
      Wr.Flush(wr);
     END;
  END Write;

PROCEDURE Read(rd :Rd.T): ReadData  =
  VAR d: ReadData;
  BEGIN
    FileFmt.ReadHeader(rd, "PZSegment.List", NewFileVersion);
    d.cmt := FileFmt.ReadComment(rd, '|');
    WITH
      nSegments = NGet.Int(rd, "segments")
     DO
      FGet.EOL(rd);
      d.s := NEW(REF List, nSegments);
      FOR i := 0 TO nSegments-1 DO
        FGet.Skip(rd, FGet.Spaces);
        WITH si = d.s[i] DO
          si.cvx := FGet.Int(rd); FGet.Skip(rd, FGet.Spaces);
          si.tot := FGet.Int(rd); FGet.Skip(rd, FGet.Spaces);
          si.ini := FGet.Int(rd); FGet.Skip(rd, FGet.Spaces);
          si.ns := FGet.Int(rd); FGet.Skip(rd, FGet.Spaces);
          WITH c = FGet.Char(rd) DO
            IF c = '-' THEN
              si.rev := TRUE
            ELSIF c = '+' THEN
              si.rev := FALSE
            ELSE
              <* ASSERT FALSE *>
            END
          END;
          FGet.EOL(rd)
        END
      END;
      FileFmt.ReadFooter(rd, "PZSegment.List");
      RETURN d;
    END;
  END Read;

PROCEDURE ReadOld(rd :Rd.T; READONLY tot: NATS): ReadData  =
  VAR d: ReadData;
  BEGIN
    FileFmt.ReadHeader(rd, "PZSegment.List", OldFileVersion);
    d.cmt := FileFmt.ReadComment(rd, '|');
    WITH
      n = NGet.Int(rd, "segments")
     DO
      FGet.EOL(rd);
      d.s := NEW(REF List, n);
      FOR i := 0 TO n-1 DO
        FGet.Skip(rd, FGet.Spaces);
        WITH si = d.s[i] DO
          si.cvx := FGet.Int(rd); FGet.Skip(rd, FGet.Spaces);
          si.tot := tot[si.cvx];
          si.ini := FGet.Int(rd); FGet.Skip(rd, FGet.Spaces);
          WITH
            fin = FGet.Int(rd)
          DO
            si.ns := (fin - si.ini) MOD si.tot + 1
          END;
          FGet.EOL(rd);
          si.rev := FALSE;
        END
      END;
      FileFmt.ReadFooter(rd, "PZSegment.List");
      RETURN d;
    END;
  END ReadOld;

PROCEDURE GetCurves(READONLY s: List): REF BOOLS =
  CONST MaxCurves = 10000;
  VAR mxc: NAT := 0; (* Max curve number in "s" *)
  BEGIN
    WITH seen = NEW(REF BOOLS, MaxCurves)^ DO
      FOR k := 0 TO LAST(seen) DO seen[k] := FALSE END;
      FOR i := 0 TO LAST(s) DO
        WITH k = s[i].cvx DO
          IF NOT seen[k] THEN
            seen[k] := TRUE;  mxc := MAX(mxc, k)
          END;
        END
      END;
      WITH rb = NEW(REF BOOLS, mxc + 1) DO
        rb^ := SUBARRAY(seen, 0, mxc + 1);
        RETURN rb
      END
    END
  END GetCurves;

PROCEDURE Print (wr: Wr.T; READONLY s: T) =
  <* FATAL Wr.Failure, Thread.Alerted *>
  BEGIN
    WITH
      fm = FLOAT(s.tot, LONG),
      pci = FLOAT(s.ini, LONG) / fm,
      fin = (s.ini + s.ns - 1) MOD s.tot,
      pcj = FLOAT(fin, LONG) / fm
    DO
      Wr.PutText(wr, "chain " & FI(s.cvx,4));
      Wr.PutText(wr, " - " & FI(s.ns,6) & " samples");
      Wr.PutText(wr, " out of " & FI(s.tot,6));
      Wr.PutText(wr, " from " & Fmt.Pad(Fmt.Int(s.ini), 4, '0'));
      Wr.PutText(wr, "(" & FLR(pci,4,3));
      Wr.PutText(wr, ") to " & Fmt.Pad(Fmt.Int(fin), 4, '0'));
      Wr.PutText(wr, "(" & FLR(pcj,4,3) & ") "  );
      IF s.rev THEN
        Wr.PutText(wr, "-")
      ELSE
        Wr.PutText(wr, "+")
      END;
    END;
  END Print;

PROCEDURE FI(x: INT; w: NAT): TEXT =
  BEGIN
    RETURN Fmt.Pad(Fmt.Int(x), w)
  END FI;

PROCEDURE FLR(x: LONG; w, p: NAT): TEXT =
  BEGIN
    RETURN Fmt.Pad(Fmt.LongReal(x, Fmt.Style.Fix, prec := p), w)
  END FLR;

BEGIN
END PZSegment.
(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
