MODULE PZSmooth;

(* Last edited on 1999-08-15 09:03:09 by hcgl *)
 
IMPORT  PZLRChain, LR3, PZGeo, PZLR3Chain, Math; 
FROM PZTypes IMPORT LONG, NAT;

(*DEBUG 
IMPORT Wr, Fmt, Thread;
FROM Stdio IMPORT stderr;
<* FATAL Wr.Failure, Thread.Alerted *>
GUBED*)

CONST 
  (* Sqrt2     = 1.4142135623730950488d0; *)
  SqrtHalf  = 0.7071067811865475244d0;
  (* SqrtPi    = 1.77245385090551603d0; *)
  SqrtTwoPi = 2.50662827463100050d0;

PROCEDURE Polygonal(
    sigma: LONG;
    READONLY p: PZLR3Chain.T; 
    READONLY t: PZLRChain.T;
    tPeriod: LONG;
    tStart: LONG;
    VAR u: PZLR3Chain.T;
    (* VAR v: PZLR3Chain.T; *)
    (* VAR w: PZLR3Chain.T; *)
  ) =
  VAR R: LONG;
  BEGIN 
    WITH 
      m = NUMBER(p),
      n = NUMBER(u),
      tStep = tPeriod / FLOAT(n, LONG) 
    DO
      <* ASSERT NUMBER(p) = NUMBER(t) *>
      (* <* ASSERT NUMBER(v) = NUMBER(u) *> *)
      (* <* ASSERT NUMBER(w) = NUMBER(u) *> *)
      IF tPeriod = 0.0d0 OR m = 1 THEN
        WITH b = PZLR3Chain.Barycenter(p) DO
          FOR i := 0 TO LAST(u) DO
            u[i] := b; 
            (* v[i] := LR3.T{0.0d0, ..}; *)
            (* w[i] := LR3.T{0.0d0, ..}; *)
          END
        END;
        RETURN 
      END;
      (*DEBUG Wr.PutText(stderr, "delta:" & FLR(delta,8,4) & "\n"); GUBED*)
      
      (* Computes length of polynomial: *)
      R := LR3.Dist(p[m-1], p[0]);
      FOR i := 1 TO LAST(p) DO R := R + LR3.Dist(p[i], p[i-1]) END;
      
      (* Clear sample arrays: *)
      FOR i := 0 TO n-1 DO 
        u[i] := p[0]; 
        (* v[i] := LR3.T{0.0d0, 0.0d0, 0.0d0}; *)
        (* w[i] := LR3.T{0.0d0, 0.0d0, 0.0d0}; *)
      END;
      
      (* Filter and sample: *)
      SplatCircus(
        sigma, 
        ta := t[0],           pa := p[0], 
        tz := t[0] + tPeriod, pz := p[0],
        p := SUBARRAY(p, 1, m-1),
        t := SUBARRAY(t, 1, m-1),
        R := R,
        tStart := tStart, tStep := tStep,
        u := u
        (* , v := v, w := w *)
      );
    END
  END Polygonal;
 
TYPE ValDiff = RECORD v (*, dv, ddv*): LONG END;

PROCEDURE SplatCircus(
    sigma: LONG;
    ta: LONG;
    READONLY pa: LR3.T;
    tz: LONG;
    READONLY pz: LR3.T;
    READONLY p: PZLR3Chain.T; 
    READONLY t: PZLRChain.T;
    R: LONG;
    tStart: LONG;
    tStep: LONG;
    VAR u: PZLR3Chain.T;
    (* VAR v: PZLR3Chain.T; *)
    (* VAR w: PZLR3Chain.T; *)
  ) =
  (*
    Filters a piecewise linear function "p(t)" that is zero for "t <=
    ta" or "t >= tz", and whose graph has vertices
    
      "(ta,0)", 
      "(t[0],d[0])", "(t[1],d[1])", ... "(t[N-1],d[N-1])",
      "(tz,0)",
    
    where "N = NUMBER(t) = NUMBER(p)", and "d[i]" is the difference
    between "p[i]" and the linear interpolation of "(ta,pa)" and
    "(tz,pz)".
    
    Evaluates the filtered function and its derivatives at equally
    spaced times "tStart + k*tStep", and adds those values to the
    vectors "u[k]", "v[k]", "w[k]".
    
    The procedure assumes that "d[i]" is at most "R".
  *)
  
  VAR dtMin: LONG;
  
  PROCEDURE DoSplatCircus(
      ta: LONG;
      READONLY pa: LR3.T;
      tz: LONG;
      READONLY pz: LR3.T;
      lo: NAT;
      hi: NAT;
    ) =
    (*
      Performs "SplatCircus" for the points "t[lo..hi], p[lo..hi]"
    *)
    BEGIN
      IF (lo > hi) OR (tz - ta < dtMin) THEN RETURN END;
      WITH
        md = (lo+hi) DIV 2,
        tm = t[md],
        pm = p[md],
        rm = PZGeo.LinearInterpolate(tm, ta, pa, tz, pz),
        dm = LR3.Sub(pm, rm)
      DO
        SplatTent(dm, ta, tm, tz, tStart, tStep, sigma, u (*, v, w *));
        IF lo < md THEN DoSplatCircus(ta, pa, tm, pm, lo, md-1) END;
        IF md < hi THEN DoSplatCircus(tm, pm, tz, pz, md+1, hi) END;
      END
    END DoSplatCircus;
  
  BEGIN
    <* ASSERT NUMBER(p) = NUMBER(t) *>
    dtMin := 1.0d-4 * SqrtTwoPi * Math.sqrt(sigma)/R;
    DoSplatCircus(ta, pa, tz, pz, 0, LAST(t))
  END SplatCircus;

PROCEDURE SplatTent(
    READONLY pi: LR3.T;
    tm: LONG;
    ti: LONG;
    tp: LONG;
    tStart: LONG;
    tStep: LONG;
    sigma: LONG;
    VAR u: PZLR3Chain.T;
    (* VAR v: PZLR3Chain.T; *)
    (* VAR w: PZLR3Chain.T; *)
  ) =
  BEGIN 
    WITH 
      fiveSigma = 5.0d0 * sigma,
      n = NUMBER(u), 
      jMin = CEILING((tm - tStart - fiveSigma)/tStep),
      jMax = FLOOR((tp - tStart + fiveSigma)/tStep)
    DO 
      FOR j:= jMin TO jMax DO 
        WITH 
          t = tStart + tStep * FLOAT(j,LONG),
          ci = EvalSmoothedTent(t-ti, tm-ti, 0.0d0, tp-ti, sigma),
          jm = j MOD n,
          uj = u[jm]
          (* , vj = v[jm], *)
          (* wj = w[jm] *)
        DO 
          FOR k := 0 TO 2 DO  
            <* ASSERT pi[k] + 1.0d0 > pi[k] *>
            uj[k] := uj[k] + pi[k] * ci.v;
            (* vj[k] := vj[k] + pi[k] * ci.dv; *)
            (* wj[k] := wj[k] + pi[k] * ci.ddv; *)
          END;
        END;
      END
    END
  END SplatTent;


PROCEDURE EvalSmoothedTent(
    t: LONG;
    tm: LONG; 
    ti: LONG; 
    tp: LONG; 
    sigma: LONG
  ): ValDiff =
  (*
    Evaluates the tent function with graph "(tm,0) (ti,1) (tp,0)",
    filtered by a Gaussian of duration "sigma", and its derivatives,
    at time "t". *)
  BEGIN
    IF tp <= tm THEN RETURN ValDiff{0.0d0 (*, 0.0d0, 0.0d0 *)} END;
    WITH
      w = tp - tm,
      (* Fix "ti" so that it does not coincide with "tm" or "tp": *)
      eps = 1.0d-5 * w,
      tiFix = MIN(MAX(ti, tm + eps), tp-eps),
      (* Now evaluate the filtered tent: *)
      a = tiFix-tm, 
      b = tp-tiFix,
      A = 1.0d0/a, 
      B = 1.0d0/b,
      AB = A+B,
      r0 = EvalSmoothedRamp(t-tm,sigma),
      r1 = EvalSmoothedRamp(t-tiFix,sigma),
      r2 = EvalSmoothedRamp(t-tp,sigma),
      v = A * r0.v - AB * r1.v + B * r2.v
      (* , *)
      (* dv = A * r0.dv - AB * r1.dv + B * r2.dv, *)
      (* ddv = A * r0.ddv - AB * r1.ddv + B * r2.ddv *)
    DO
      <* ASSERT v + 1.0d0 > v *>
      RETURN ValDiff{v (*, dv, ddv *)}
    END
  END EvalSmoothedTent;


PROCEDURE EvalSmoothedRamp(t: LONG; sigma: LONG): ValDiff =
  (*
    Evaluates the unit ramp function "MAX(0,x)", 
    filtered by a Gaussian of duration "sigma", 
    and its derivatives, at time "t". *)
  BEGIN
    WITH 
      ts = t/sigma,
      ts2 = ts*ts,
      er1 = 1.0d0 + Math.erf(SqrtHalf*ts),
      ga = Math.exp(-0.5d0*ts2),
      v = 0.5d0 * t * er1 + (sigma/SqrtTwoPi) * ga
      (* , dv = 0.5d0 * er1, *)
      (* ddv = ga/(sigma*SqrtTwoPi) *)
    DO 
      <* ASSERT v + 1.0d0 > v *>
      RETURN ValDiff{v (*, dv, ddv *)}
    END
  END EvalSmoothedRamp;

(*DEBUG
PROCEDURE FLR(x: LONG; w, p: NAT): TEXT =
  BEGIN
    RETURN Fmt.Pad(Fmt.LongReal(x, Fmt.Style.Fix, prec := p), w)
  END FLR;
GUBED*)
  
BEGIN
END PZSmooth.


(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
