MODULE PZSymbolChain;

(* Last edited on 1999-11-15 00:05:18 by hcgl *)

IMPORT Rd, Wr, Thread, FileFmt, NPut, NGet, FPut, FGet, Fmt, OSError, FileRd;
IMPORT PZMatch, PZProc, PZSegment, PZSymbol;
FROM Math IMPORT sqrt;
FROM Stdio IMPORT stderr;
FROM PZTypes IMPORT LONG, LONGS, NAT, BOOL, BOOLS;

PROCEDURE Trim(READONLY c: T; start, length: NAT): REF T =
  BEGIN
    WITH rx = NEW(REF T, length), x = rx^ DO
      DoTrim(c, start, length, x);
      RETURN rx
    END;
  END Trim;

PROCEDURE DoTrim(READONLY c: T; start, length: NAT; VAR x: T) =
  BEGIN
    WITH
      N = NUMBER(c),
      ini = start MOD N,
      fin = ini + length - 1
    DO
      IF fin < N  THEN
        x := SUBARRAY(c, ini, length)
      ELSE
        SUBARRAY(x, 0, N-ini) := SUBARRAY(c, ini, N-ini);
        SUBARRAY(x, N-ini, (fin+1) MOD N):= SUBARRAY(c, 0, (fin+1) MOD N);
      END;
    END;
  END DoTrim;

PROCEDURE ReverseAndComplement(VAR c: T) =
  VAR aux: PZSymbol.T;
  BEGIN
    WITH N = NUMBER(c), L = N-1 DO
      FOR i := 0 TO (N DIV 2)-1  DO
        aux := c[i]; c[i] := c[L-i]; c[L-i] := aux
      END;
      FOR i := 0 TO N-1 DO c[i] := PZSymbol.Complement(c[i]) END
    END;
  END ReverseAndComplement;

PROCEDURE ExtractSegment(READONLY c: T; s: PZSegment.T): REF T =
  BEGIN
    <* ASSERT NUMBER(c) = s.tot *>
    WITH rx = NEW(REF T, s.ns), x = rx^ DO
      DoExtractSegment(c, s, x);
      RETURN rx
    END
  END ExtractSegment;

PROCEDURE DoExtractSegment(READONLY c: T; s: PZSegment.T; VAR x: T) =
  BEGIN
    <* ASSERT NUMBER(c) = s.tot *>
    DoTrim(c, s.ini, s.ns, x);
    IF s.rev THEN
      ReverseAndComplement(x);
    END;
  END DoExtractSegment;

PROCEDURE PrintSegment(wr: Wr.T; READONLY c: T; s: PZSegment.T) =
  <* FATAL Thread.Alerted, Wr.Failure *>
  BEGIN
    WITH
      m = NUMBER(c)
    DO
      <* ASSERT NUMBER(c) = s.tot *>
      IF s.rev THEN
        FOR r := s.ns-1 TO 0 BY -1 DO
          WITH ch = c[(s.ini + r) MOD m] DO
            Wr.PutText(wr, Fmt.Char(PZSymbol.Complement(ch)))
          END;
        END;
      ELSE
        FOR r := 0 TO s.ns-1  DO
          WITH ch = c[(s.ini + r) MOD m] DO
            Wr.PutText(wr, Fmt.Char(ch))
          END;
        END;
        Wr.PutText(wr, "\n");
      END
    END;
  END PrintSegment;

CONST 
  FileVersion = "99-08-22";

PROCEDURE Write(
    wr :Wr.T;
    cmt: TEXT;
    READONLY c: T;
    sigma: LONG;
  ) =
   <* FATAL Thread.Alerted, Wr.Failure *>
  BEGIN
    WITH
      n = LAST(c)
    DO
      FileFmt.WriteHeader(wr, "PZSymbolChain.T", FileVersion);
      FileFmt.WriteComment(wr, cmt, '|');
      NPut.Int(wr, "samples", NUMBER(c)); FPut.EOL(wr);
      NPut.LongReal(wr, "sigma", sigma); FPut.EOL(wr);
      FOR i := 0 TO n  DO
        FPut.Char(wr, c[i]); FPut.EOL(wr)
      END;
      FPut.EOL(wr);
      FileFmt.WriteFooter(wr, "PZSymbolChain.T");
      Wr.Flush(wr);
     END;
  END Write;

PROCEDURE Read(rd: Rd.T; headerOnly: BOOL): ReadData  =
  VAR d: ReadData;
  BEGIN
    FileFmt.ReadHeader(rd,"PZSymbolChain.T", FileVersion);
    d.cmt := FileFmt.ReadComment(rd, '|');
    d.samples := NGet.Int(rd, "samples"); FGet.EOL(rd);
    d.sigma := NGet.LongReal(rd, "sigma"); FGet.EOL(rd);
    WITH
      n = d.samples
    DO
      IF headerOnly THEN
        d.c := NIL
      ELSE
        d.c := NEW(REF T, n);
        FOR i := 0 TO n-1 DO
          FGet.Skip(rd, FGet.Blanks);
          d.c[i] := FGet.Char(rd);
        END;
        FileFmt.ReadFooter(rd, "PZSymbolChain.T");
      END;
      RETURN d;
    END;
  END Read;

CONST
  OldFileVersion = "97-02-03";

PROCEDURE ReadOld(rd: Rd.T; sigma: LONG): ReadData  =
  VAR d: ReadData;
      length, lambda, epsilon, delta: LONG;
  BEGIN
    FileFmt.ReadHeader(rd,"PZSymbolChain.T", OldFileVersion);
    d.cmt := FileFmt.ReadComment(rd, '|');
    d.samples := NGet.Int(rd, "samples"); FGet.EOL(rd);
    WITH
      n = d.samples
     DO
      length := NGet.LongReal(rd, "length"); FGet.EOL(rd);
      epsilon := NGet.LongReal(rd, "epsilon"); FGet.EOL(rd);
      delta := NGet.LongReal(rd, "delta"); FGet.EOL(rd);
      d.c := NEW(REF T, n);
      d.sigma := sigma;
      FOR i := 0 TO n-1 DO
        FGet.Skip(rd, FGet.Blanks);
        WITH 
          c = FGet.Char(rd),
          x = FLOAT(PZSymbol.ToInt(c), LONG)
        DO
          d.c[i] := PZSymbol.Encode(PZProc.OldStretch(x, epsilon, delta), lambda)
        END
      END;
      FileFmt.ReadFooter(rd, "PZSymbolChain.T");
      RETURN d;
    END;
  END ReadOld;

PROCEDURE ReadAll(
    prefix: TEXT;
    band: NAT;
    extension: TEXT;
    READONLY sel: BOOLS;
    headerOnly: BOOL;
    dir: TEXT := ".";
  ): ReadAllData =
  <* FATAL Wr.Failure, Thread.Alerted, Rd.Failure *>
  VAR sigmaMin: LONG := LAST(LONG);
      sigmaMax: LONG := FIRST(LONG);
      cmt: TEXT := ""; 
  CONST NoData = ReadData{ c := NIL, samples := 0, sigma := 0.0d0, cmt := "NOT READ" };
  BEGIN
    Wr.PutText(stderr, "reading symbolic chains:\n");
    WITH
      rr = NEW(REF ARRAY OF ReadData, NUMBER(sel)), r = rr^
    DO
      FOR k := 0 TO LAST(sel) DO
        IF sel[k] THEN
          Wr.PutText(stderr, "  " & Fmt.Pad(Fmt.Int(k), 4, '0') & ": ");
          WITH
            chainTag  = Fmt.Pad(Fmt.Int(k), 4, '0'),
            bandTag = Fmt.Pad(Fmt.Int(band), 3, '0'),
            fileName = prefix & bandTag & extension,
            fullName = dir & "/" & chainTag &"/"& fileName,
            rd = OpenRead(fullName)
          DO
            IF rd # NIL THEN
              Wr.PutText(stderr, "read from " & fullName & "\n");
              WITH data = Read(rd, headerOnly := headerOnly) DO
                r[k] := data;
                <* ASSERT data.sigma >= 0.0d0 *>
                IF sigmaMin > sigmaMax THEN 
                  cmt := data.cmt
                END;
                sigmaMin := MIN(sigmaMin, data.sigma);
                sigmaMax := MAX(sigmaMax, data.sigma);
              END;
              Rd.Close(rd)
            ELSE
              r[k] := NoData;
              Wr.PutText(stderr, "file " & fullName & " not found.\n");
            END
          END
        ELSE
          r[k] := NoData;
        END
      END;
      RETURN ReadAllData {
        chain := rr,
        sigmaMin := sigmaMin,
        sigmaMax := sigmaMax,
        cmt := 
          "Chains read from " & 
          dir & "/NNNN/" & prefix & "BBB" & extension & "\n" &
          cmt
      }
    END
  END ReadAll;

PROCEDURE OpenRead(name: TEXT): Rd.T =
  BEGIN
    TRY
      RETURN FileRd.Open(name)
    EXCEPT
      OSError.E => RETURN NIL
    END
  END OpenRead;

PROCEDURE FindBestMatch(
    READONLY a: T;
    READONLY b: T;
    READONLY decode: ARRAY PZSymbol.T OF LONG;
    READONLY errorVar: ARRAY PZSymbol.T OF LONG;
    step: LONG;
    maxDistSqr: LONG;
    skipDistSqr: LONG;
    removeUnmatchedEnds: BOOL;
    VAR (*OUT*) mt: REF PZMatch.T;
    VAR (*OUT*) avgDisc: LONG;
    VAR (*OUT*) length: LONG;
    VAR (*OUT*) matchedLength: LONG;
    VAR (*WORK*) rCost: REF PZMatch.CostMatrix;
  ) =

  TYPE Pair = PZMatch.Pair;

  PROCEDURE StepDistSqr(READONLY p, q: Pair): LONG =
    (*
      Computes the mean squared chain distance for one matching step "p -> q": *)
    BEGIN
      WITH
        mdSqr = PZSymbol.IntgDistSqr(
          a[p[0]], a[q[0]], b[p[1]], b[q[1]],
          complement := FALSE,
          decode := decode, 
          errorVar := errorVar
        )
      DO
        RETURN MIN(mdSqr, maxDistSqr);
      END
    END StepDistSqr;

  PROCEDURE StepIntgDiscSqr(READONLY p, q: Pair): LONG =
    (*
      Computes the integrated discrepancy squared
      for one matching step "p -> q": *)
    BEGIN
      WITH
        d2 = StepDistSqr(p, q)
      DO
        IF p[0] = q[0] OR p[1] = q[1] THEN
          RETURN step * (d2 + skipDistSqr)/2.0d0
        ELSE
          RETURN step * d2
        END
      END
    END StepIntgDiscSqr;

  PROCEDURE StepIsMatched(READONLY p, q: Pair): BOOL =
    (*
      A "PZMatch.StepPredicate" that returns TRUE if the step "p->q"
      has average chain distance less than "maxDist". *)
    BEGIN
      RETURN StepDistSqr(p, q) < maxDistSqr
    END StepIsMatched;

  VAR intgDiscSqr: LONG;
  BEGIN
    WITH
      NA = NUMBER(a),
      NB = NUMBER(b)
    DO
      PZMatch.FindBest(
        NA, NB,
        stepCostFn := StepIntgDiscSqr,
        (*OUT*) mt := mt, totCost := intgDiscSqr,
        (*WORK*) rCost := rCost
      );
      IF removeUnmatchedEnds THEN
        mt := PZMatch.RemoveUnmatchedEnds(mt^, StepIsMatched);
        intgDiscSqr := PZMatch.Sum(mt^, stepFn := StepIntgDiscSqr)
      END;
      <* ASSERT NUMBER(mt^) > 0 *>
      WITH
        m = NUMBER(mt^),
        na = mt[m-1][0] - mt[0][0],
        nb = mt[m-1][1] - mt[0][1]
      DO
        length := step * FLOAT(na + nb, LONG)/2.0d0;
        IF length <= 0.0d0 THEN
          avgDisc := sqrt(maxDistSqr)
        ELSE
          avgDisc := sqrt(intgDiscSqr/length)
        END
      END;
      WITH msc = PZMatch.MatchedStepCount(mt^, StepIsMatched) DO
        matchedLength := step * FLOAT(msc, LONG)/2.0d0
      END
    END;
  END FindBestMatch;

PROCEDURE RefineMatch(
    READONLY a: T;
    READONLY b: T;
    READONLY decode: ARRAY PZSymbol.T OF LONG;
    READONLY errorVar: ARRAY PZSymbol.T OF LONG;
    step: LONG;
    READONLY mtOld: PZMatch.T;
    maxAdjust: NAT;
    critDistSqr: LONG;
    maxDistSqr: LONG;
    skipDistSqr: LONG;
    VAR (*OUT*) mt: REF PZMatch.T;
    VAR (*OUT*) mismatch: LONG;
    VAR (*OUT*) length: LONG;
    VAR (*OUT*) matchedLength: LONG;
    VAR (*OUT*) minAvgDisc: LONGS;
    VAR (*WORK*) rCost: REF PZMatch.CostMatrix;
  ) =

  TYPE Pair = PZMatch.Pair;
  
  PROCEDURE StepDistSqr (READONLY p, q: Pair): LONG =
    (*
      The mean square distance between the chain steps
      "a[p[0]]->a[q[0]]" and "b[p[1]]->b[q[1]]". *)
    BEGIN
      WITH
        mdSqr = PZSymbol.IntgDistSqr(
          a[p[0]], a[q[0]], b[p[1]], b[q[1]],
          complement := FALSE,
          decode := decode, 
          errorVar := errorVar
        )
      DO
(*      Wr.PutText(stderr,"a[" & Fmt.Int(p[0]) & ", "& Fmt.Int(q[0]) & "]:" 
         & Fmt.Char(a[p[0]]) & "  " &  Fmt.Char(a[q[0]]) & "  ");  
        Wr.PutText(stderr,"b[" & Fmt.Int(p[1]) & ", " & Fmt.Int(q[1]) & "]:" 
        & Fmt.Char(b[p[1]]) & "  " &  Fmt.Char(b[q[1]]) & "   ");
        Wr.PutText(stderr,"mdSqr: " & Fmt.LongReal(mdSqr) & "   ");
        Wr.PutText(stderr,"mdSqr-critDistSqr: " 
        & Fmt.LongReal(mdSqr-critDistSqr) & "\n");
 *)        
        RETURN MIN(mdSqr, maxDistSqr)
      END
    END StepDistSqr;

  PROCEDURE StepMismatch (READONLY p, q: Pair): LONG =
    (*
      A "PZMatch.StepFunction" whose value is the mismatch of the step
      "p -> q", i.e. the integrated square distance between the chain
      steps "a[p[0]]->a[q[0]]" and "b[p[1]]->b[q[1]]", plus the skip
      distance squared, minus the critical distance squared,
      everything times the step length. *)
    BEGIN
      WITH
        d2 = StepDistSqr(p, q) - critDistSqr,
        hs2 = step * (d2 + skipDistSqr)/2.0d0,
        s2 = step * d2
      DO
        IF p[0] = q[0] OR p[1] = q[1] THEN
          RETURN hs2
        ELSE
          RETURN s2
        END
      END
    END StepMismatch;

  PROCEDURE StepIsMatched(READONLY p, q: Pair): BOOL =
    (*
      A "PZMatch.StepPredicate" that returns TRUE if the step "p->q"
      has average chain distance less than "maxDist". *)
    BEGIN
      RETURN StepDistSqr(p, q) < maxDistSqr
    END StepIsMatched;

  BEGIN
    WITH
      NA = NUMBER(a),
      NB = NUMBER(b)
    DO
      <* ASSERT NA > 0 *>
      <* ASSERT NB > 0 *>
      PZMatch.Refine(
        NA := NA, NB := NB,
        stepCostFn := StepMismatch,
        mtOld := mtOld,
        maxAdjust := maxAdjust,
        (*OUT*) mt := mt,
        (*OUT*) totCost := mismatch,
        (*OUT*) minTotCost := minAvgDisc,
        (*WORK*) rCost := rCost
      );
      <* ASSERT  NUMBER(mt^) > 0 *>
      WITH
        mt = mt^,
        m = NUMBER(mt),
        stepCount = (mt[m-1][0] - mt[0][0]) + (mt[m-1][1] - mt[0][1]),
        matchedStepCount = PZMatch.MatchedStepCount(mt, StepIsMatched)
      DO
        length := FLOAT(stepCount, LONG) * step / 2.0d0;
        matchedLength := step * FLOAT(matchedStepCount, LONG)/2.0d0
      END;
      (* Compute the minimum average distance for each matching length "K". *)
      (* Note that "PZMatch.Refine" returns total mismatch, not just "intgDiscSqr" *)
      <* ASSERT  minAvgDisc[0] = 0.0d0 *>
      FOR K := 1 TO LAST(minAvgDisc) DO
        WITH Lk = step*FLOAT(K, LONG)/2.0d0 DO
          minAvgDisc[K] := sqrt(MAX(0.0d0, minAvgDisc[K]/Lk + critDistSqr))
        END
      END
    END;
  END RefineMatch;

BEGIN
END PZSymbolChain.
(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
