MODULE QPut;

IMPORT Wr, Fmt, FmtBuf;
IMPORT Real AS R, LongReal AS LR;
FROM Thread IMPORT Alerted;
FROM Wr IMPORT Failure;

PROCEDURE EOL(wr: Wr.T) =
  <* FATAL Failure, Alerted *>
  BEGIN Wr.PutChar(wr, '\n') END EOL;

PROCEDURE Colon(wr: Wr.T) =
  <* FATAL Failure, Alerted *>
  BEGIN Wr.PutChar(wr, ':'); Wr.PutChar(wr, ' ') END Colon;

PROCEDURE Space(wr: Wr.T; n: CARDINAL) =
  <* FATAL Failure, Alerted *>
  BEGIN FOR i := 1 TO n DO Wr.PutChar(wr, ' ') END END Space;

PROCEDURE Char(wr: Wr.T; x: CHAR) =
  <* FATAL Failure, Alerted *>
  BEGIN 
    Wr.PutChar(wr, x)
  END Char;

PROCEDURE Bool(wr: Wr.T; x: BOOLEAN) =
  <* FATAL Failure, Alerted *>
  BEGIN 
    IF x THEN Wr.PutChar(wr, 'T') ELSE Wr.PutChar(wr, 'F') END
  END Bool;

PROCEDURE Int(wr: Wr.T; x: INTEGER; pad: CARDINAL := 0) =
  <* FATAL Failure, Alerted *>
  CONST MaxIntDigits = 30; (* Guess... *)
  VAR buf: ARRAY [0 .. MaxIntDigits] OF CHAR;
  BEGIN
    WITH n = FmtBuf.Int(buf, x) DO
      WHILE pad > n DO Wr.PutChar(wr, ' '); DEC(pad) END;
      Wr.PutString(wr, SUBARRAY(buf, 0, n))
    END
  END Int;

PROCEDURE Real(
    wr: Wr.T; x: REAL; 
    prec: [1..RMaxPrec] := RMaxPrec; 
    style: Fmt.Style := Fmt.Style.Auto;
  ) =
  <* FATAL Failure, Alerted *>
  VAR buf: ARRAY [0 .. RMaxPrec + R.MaxExpDigits + 5] OF CHAR;
  BEGIN
    WITH n = FmtBuf.Real(buf, x, prec := prec, style := style) DO
      Wr.PutString(wr, SUBARRAY(buf, 0, n))
    END
  END Real;

PROCEDURE LongReal(
    wr: Wr.T; x: LONGREAL; 
    prec: [1..LRMaxPrec] := LRMaxPrec; 
    style: Fmt.Style := Fmt.Style.Auto;
  ) =
  <* FATAL Failure, Alerted *>
  VAR buf: ARRAY [0 .. LRMaxPrec + LR.MaxExpDigits + 5] OF CHAR;
  BEGIN
    WITH n = FmtBuf.LongReal(buf, x, prec := prec, style := style) DO
      Wr.PutString(wr, SUBARRAY(buf, 0, n))
    END
  END LongReal;

BEGIN
END QPut.
(*
  Copyright  2001 Universidade Estadual de Campinas (UNICAMP).
  Authors: Helena C. G. Leito and Jorge Stolfi.
  
  This file can be freely distributed, used, and modified, provided
  that this copyright and authorship notice is preserved, and that any
  modified versions are clearly marked as such.
  
  This software has NO WARRANTY of correctness or applicability for
  any purpose. Neither the authors nor their employers chall be held
  responsible for any losses or damages that may result from its use.
*)
